#
# This file is licensed under the Affero General Public License (AGPL) version 3.
#
# Copyright (C) 2025 New Vector, Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# See the GNU Affero General Public License for more details:
# <https://www.gnu.org/licenses/agpl-3.0.html>.
import logging
import random
from dataclasses import dataclass
from typing import TYPE_CHECKING, Collection, cast

from twisted.internet.defer import Deferred

from synapse.events import EventBase
from synapse.replication.tcp.streams._base import StickyEventsStream
from synapse.storage.database import (
    DatabasePool,
    LoggingDatabaseConnection,
    LoggingTransaction,
    make_in_list_sql_clause,
)
from synapse.storage.databases.main.cache import CacheInvalidationWorkerStore
from synapse.storage.databases.main.state import StateGroupWorkerStore
from synapse.storage.engines import PostgresEngine, Sqlite3Engine
from synapse.storage.util.id_generators import MultiWriterIdGenerator
from synapse.util.duration import Duration

if TYPE_CHECKING:
    from synapse.server import HomeServer

logger = logging.getLogger(__name__)

DELETE_EXPIRED_STICKY_EVENTS_INTERVAL = Duration(hours=1)
"""
Remove entries from the sticky_events table at this frequency.
Note: don't be misled, we still honour shorter expiration timeouts,
because readers of the sticky_events table filter out expired sticky events
themselves, even if they aren't deleted from the table yet.

Currently just an arbitrary choice.
Frequent enough to clean up expired sticky events promptly,
especially given the short cap on the lifetime of sticky events.
"""


@dataclass(frozen=True)
class StickyEventUpdate:
    stream_id: int
    room_id: str
    event_id: str
    soft_failed: bool


class StickyEventsWorkerStore(StateGroupWorkerStore, CacheInvalidationWorkerStore):
    def __init__(
        self,
        database: DatabasePool,
        db_conn: LoggingDatabaseConnection,
        hs: "HomeServer",
    ):
        super().__init__(database, db_conn, hs)

        self._can_write_to_sticky_events = (
            self._instance_name in hs.config.worker.writers.events
        )

        # Technically this means we will cleanup N times, once per event persister, maybe put on master?
        if self._can_write_to_sticky_events:
            # Start a looping call to clean up the `sticky_events` table
            #
            # Because this will run once per event persister (for now),
            # randomly stagger the initial time so that they don't all
            # coincide with each other if the workers are deployed at the
            # same time. This allows each cleanup to be somewhat more effective
            # than if they all started at the same time, as they would all be
            # cleaning up the same thing whereas each worker gets to clean up a little
            # throughout the hour when they're staggered.
            #
            # Concurrent execution of the same deletions could also lead to
            # repeatable serialisation violations in the database transaction,
            # meaning we'd have to retry the transaction several times.
            #
            # This staggering is not critical, it's just best-effort.
            self.clock.call_later(
                # random() is 0.0 to 1.0
                DELETE_EXPIRED_STICKY_EVENTS_INTERVAL * random.random(),
                self.clock.looping_call,
                self._run_background_cleanup,
                DELETE_EXPIRED_STICKY_EVENTS_INTERVAL,
            )

        self._sticky_events_id_gen: MultiWriterIdGenerator = MultiWriterIdGenerator(
            db_conn=db_conn,
            db=database,
            notifier=hs.get_replication_notifier(),
            stream_name="sticky_events",
            server_name=self.server_name,
            instance_name=self._instance_name,
            tables=[
                ("sticky_events", "instance_name", "stream_id"),
            ],
            sequence_name="sticky_events_sequence",
            writers=hs.config.worker.writers.events,
        )

        if hs.config.experimental.msc4354_enabled and isinstance(
            self.database_engine, Sqlite3Engine
        ):
            import sqlite3

            if sqlite3.sqlite_version_info < (3, 40, 0):
                raise RuntimeError(
                    f"Experimental MSC4354 Sticky Events enabled but SQLite3 version is too old: {sqlite3.sqlite_version_info}, must be at least 3.40. Disable MSC4354 Sticky Events, switch to Postgres, or upgrade SQLite. See https://github.com/element-hq/synapse/issues/19428"
                )

    def process_replication_position(
        self, stream_name: str, instance_name: str, token: int
    ) -> None:
        if stream_name == StickyEventsStream.NAME:
            self._sticky_events_id_gen.advance(instance_name, token)
        super().process_replication_position(stream_name, instance_name, token)

    def get_max_sticky_events_stream_id(self) -> int:
        """Get the current maximum stream_id for thread subscriptions.

        Returns:
            The maximum stream_id
        """
        return self._sticky_events_id_gen.get_current_token()

    def get_sticky_events_stream_id_generator(self) -> MultiWriterIdGenerator:
        return self._sticky_events_id_gen

    async def get_sticky_events_in_rooms(
        self,
        room_ids: Collection[str],
        *,
        from_id: int,
        to_id: int,
        now: int,
        limit: int | None,
    ) -> tuple[int, dict[str, list[str]]]:
        """
        Fetch all the sticky events' IDs in the given rooms, with sticky stream IDs satisfying
        from_id < sticky stream ID <= to_id.

        The events are returned ordered by the sticky events stream.

        Args:
            room_ids: The room IDs to return sticky events in.
            from_id: The sticky stream ID that sticky events should be returned from (exclusive).
            to_id: The sticky stream ID that sticky events should end at (inclusive).
            now: The current time in unix millis, used for skipping expired events.
            limit: Max sticky events to return, or None to apply no limit.
        Returns:
            to_id, dict[room_id, list[event_ids]]
        """
        sticky_events_rows = await self.db_pool.runInteraction(
            "get_sticky_events_in_rooms",
            self._get_sticky_events_in_rooms_txn,
            room_ids,
            from_id=from_id,
            to_id=to_id,
            now=now,
            limit=limit,
        )

        if not sticky_events_rows:
            return to_id, {}

        # Get stream_id of the last row, which is the highest
        new_to_id, _, _ = sticky_events_rows[-1]

        # room ID -> event IDs
        room_id_to_event_ids: dict[str, list[str]] = {}
        for _, room_id, event_id in sticky_events_rows:
            events = room_id_to_event_ids.setdefault(room_id, [])
            events.append(event_id)

        return (new_to_id, room_id_to_event_ids)

    def _get_sticky_events_in_rooms_txn(
        self,
        txn: LoggingTransaction,
        room_ids: Collection[str],
        *,
        from_id: int,
        to_id: int,
        now: int,
        limit: int | None,
    ) -> list[tuple[int, str, str]]:
        if len(room_ids) == 0:
            return []
        room_id_in_list_clause, room_id_in_list_values = make_in_list_sql_clause(
            txn.database_engine, "se.room_id", room_ids
        )
        limit_clause = ""
        limit_params: tuple[int, ...] = ()
        if limit is not None:
            limit_clause = "LIMIT ?"
            limit_params = (limit,)

        if isinstance(self.database_engine, PostgresEngine):
            expr_soft_failed = "COALESCE(((ej.internal_metadata::jsonb)->>'soft_failed')::boolean, FALSE)"
        else:
            expr_soft_failed = "COALESCE(ej.internal_metadata->>'soft_failed', FALSE)"

        txn.execute(
            f"""
            SELECT se.stream_id, se.room_id, event_id
            FROM sticky_events se
            INNER JOIN event_json ej USING (event_id)
            WHERE
                NOT {expr_soft_failed}
                AND ? < expires_at
                AND ? < stream_id
                AND stream_id <= ?
                AND {room_id_in_list_clause}
            ORDER BY stream_id ASC
            {limit_clause}
            """,
            (now, from_id, to_id, *room_id_in_list_values, *limit_params),
        )
        return cast(list[tuple[int, str, str]], txn.fetchall())

    async def get_updated_sticky_events(
        self, *, from_id: int, to_id: int, limit: int
    ) -> list[StickyEventUpdate]:
        """Get updates to sticky events between two stream IDs.

        Bounds: from_id < ... <= to_id

        Args:
            from_id: The starting stream ID (exclusive)
            to_id: The ending stream ID (inclusive)
            limit: The maximum number of rows to return

        Returns:
            list of StickyEventUpdate update rows
        """

        if not self.hs.config.experimental.msc4354_enabled:
            # We need to prevent `_get_updated_sticky_events_txn`
            # from running when MSC4354 is turned off, because the query used
            # for SQLite is not compatible with Ubuntu 22.04 (as used in our CI olddeps run).
            # It's technically out of support.
            # See: https://github.com/element-hq/synapse/issues/19428
            return []

        return await self.db_pool.runInteraction(
            "get_updated_sticky_events",
            self._get_updated_sticky_events_txn,
            from_id,
            to_id,
            limit,
        )

    def _get_updated_sticky_events_txn(
        self, txn: LoggingTransaction, from_id: int, to_id: int, limit: int
    ) -> list[StickyEventUpdate]:
        if isinstance(self.database_engine, PostgresEngine):
            expr_soft_failed = "COALESCE(((ej.internal_metadata::jsonb)->>'soft_failed')::boolean, FALSE)"
        else:
            expr_soft_failed = "COALESCE(ej.internal_metadata->>'soft_failed', FALSE)"

        txn.execute(
            f"""
            SELECT se.stream_id, se.room_id, se.event_id,
            {expr_soft_failed} AS "soft_failed"
            FROM sticky_events se
            INNER JOIN event_json ej USING (event_id)
            WHERE ? < stream_id AND stream_id <= ?
            LIMIT ?
            """,
            (from_id, to_id, limit),
        )

        return [
            StickyEventUpdate(
                stream_id=stream_id,
                room_id=room_id,
                event_id=event_id,
                soft_failed=bool(soft_failed),
            )
            for stream_id, room_id, event_id, soft_failed in txn
        ]

    def insert_sticky_events_txn(
        self,
        txn: LoggingTransaction,
        events: list[EventBase],
    ) -> None:
        """
        Insert events into the sticky_events table.

        Skips inserting events:
            - if they are considered spammy by the policy server;
              (unsure if correct, track: https://github.com/matrix-org/matrix-spec-proposals/pull/4354#discussion_r2727593350)
            - if they are considered spammy by a Synapse spam checker module;
            - if they are rejected;
            - if they are outliers (they should be reconsidered for insertion when de-outliered); or
            - if they are not sticky (e.g. if the stickiness expired).

        Note: Soft-failed sticky events ARE inserted, as their soft-failed status
            could be re-evaluated later.

        Skipping the insertion of these types of 'invalid' events is useful for performance reasons because
        they would fill up the table yet we wouldn't show them to clients anyway.

        Since syncing clients can't (easily?) 'skip over' sticky events (due to being in-order, reliably delivered),
        tracking loads of invalid events in the table could make it expensive for servers to retrieve the sticky events that are actually valid.

        For instance, someone spamming 1000s of rejected or 'policy_server_spammy' events could clog up this table in a way that means we either
        have to deliver empty payloads to syncing clients, or consider substantially more than 100 events in order to gather a 100-sized batch to send down.
        """

        now_ms = self.clock.time_msec()
        # event, expires_at
        sticky_events: list[tuple[EventBase, int]] = []
        for ev in events:
            # MSC: Note: policy servers and other similar antispam techniques still apply to these events.
            # We don't filter out soft-failed events altogether (in case they get re-evaluated later),
            # so filter out `spam_checker_spammy` events specifically as we don't want to re-evaluate _those_ later.
            if (
                ev.internal_metadata.policy_server_spammy
                or ev.internal_metadata.spam_checker_spammy
            ):
                continue
            # We shouldn't be passed rejected events, but if we do, we filter them out too.
            if ev.rejected_reason is not None:
                continue
            # We can't persist outlier sticky events as we don't know the room state at that event
            if ev.internal_metadata.is_outlier():
                continue
            sticky_duration = ev.sticky_duration()
            if sticky_duration is None:
                continue
            # Calculate the end time as start_time + effective sticky duration
            expires_at = min(ev.origin_server_ts, now_ms) + sticky_duration.as_millis()
            # Filter out already expired sticky events
            if expires_at <= now_ms:
                continue

            sticky_events.append((ev, expires_at))

        if len(sticky_events) == 0:
            return

        logger.info(
            "inserting %d sticky events in room %s",
            len(sticky_events),
            sticky_events[0][0].room_id,
        )

        # Generate stream_ids in one go
        sticky_events_with_ids = zip(
            sticky_events,
            self._sticky_events_id_gen.get_next_mult_txn(txn, len(sticky_events)),
            strict=True,
        )

        self.db_pool.simple_insert_many_txn(
            txn,
            "sticky_events",
            keys=(
                "instance_name",
                "stream_id",
                "room_id",
                "event_id",
                "event_stream_ordering",
                "sender",
                "expires_at",
            ),
            values=[
                (
                    self._instance_name,
                    stream_id,
                    ev.room_id,
                    ev.event_id,
                    ev.internal_metadata.stream_ordering,
                    ev.sender,
                    expires_at,
                )
                for (ev, expires_at), stream_id in sticky_events_with_ids
            ],
        )

    async def _delete_expired_sticky_events(self) -> None:
        await self.db_pool.runInteraction(
            "_delete_expired_sticky_events",
            self._delete_expired_sticky_events_txn,
            self.clock.time_msec(),
        )

    def _delete_expired_sticky_events_txn(
        self, txn: LoggingTransaction, now: int
    ) -> None:
        """
        From the `sticky_events` table, deletes all entries whose expiry is in the past
        (older than `now`).

        This is fine because we don't consider the events as sticky anymore when that's
        happened.
        """
        txn.execute(
            """
            DELETE FROM sticky_events WHERE expires_at < ?
            """,
            (now,),
        )

    def _run_background_cleanup(self) -> Deferred:
        return self.hs.run_as_background_process(
            "delete_expired_sticky_events",
            self._delete_expired_sticky_events,
        )



#
# This file is licensed under the Affero General Public License (AGPL) version 3.
#
# Copyright 2019 The Matrix.org Foundation C.I.C.
# Copyright 2017 Vector Creations Ltd
# Copyright 2014-2016 OpenMarket Ltd
# Copyright (C) 2023 New Vector, Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# See the GNU Affero General Public License for more details:
# <https://www.gnu.org/licenses/agpl-3.0.html>.
#
# Originally licensed under the Apache License, Version 2.0:
# <http://www.apache.org/licenses/LICENSE-2.0>.
#
# [This file includes modifications made by New Vector Limited]
#
#

"""This module is responsible for getting events from the DB for pagination
and event streaming.

The order it returns events in depend on whether we are streaming forwards or
are paginating backwards. We do this because we want to handle out of order
messages nicely, while still returning them in the correct order when we
paginate bacwards.

This is implemented by keeping two ordering columns: stream_ordering and
topological_ordering. Stream ordering is basically insertion/received order
(except for events from backfill requests). The topological_ordering is a
weak ordering of events based on the pdu graph.

This means that we have to have two different types of tokens, depending on
what sort order was used:
    - stream tokens are of the form: "s%d", which maps directly to the column
    - topological tokems: "t%d-%d", where the integers map to the topological
      and stream ordering columns respectively.
"""

import logging
from typing import (
    TYPE_CHECKING,
    AbstractSet,
    Any,
    Collection,
    Iterable,
    Literal,
    Mapping,
    Protocol,
    cast,
    overload,
)

import attr
from typing_extensions import assert_never

from twisted.internet import defer

from synapse.api.constants import Direction, EventTypes, Membership
from synapse.api.filtering import Filter
from synapse.events import EventBase
from synapse.logging.context import make_deferred_yieldable, run_in_background
from synapse.logging.opentracing import tag_args, trace
from synapse.storage._base import SQLBaseStore
from synapse.storage.database import (
    DatabasePool,
    LoggingDatabaseConnection,
    LoggingTransaction,
    make_in_list_sql_clause,
)
from synapse.storage.databases.main.events_worker import EventsWorkerStore
from synapse.storage.engines import BaseDatabaseEngine, PostgresEngine, Sqlite3Engine
from synapse.storage.roommember import RoomsForUserStateReset
from synapse.storage.util.id_generators import MultiWriterIdGenerator
from synapse.types import PersistedEventPosition, RoomStreamToken, StrCollection
from synapse.util.caches.descriptors import cached, cachedList
from synapse.util.caches.stream_change_cache import StreamChangeCache
from synapse.util.cancellation import cancellable
from synapse.util.iterutils import batch_iter

if TYPE_CHECKING:
    from synapse.server import HomeServer

logger = logging.getLogger(__name__)


MAX_STREAM_SIZE = 1000


_STREAM_TOKEN = "stream"
_TOPOLOGICAL_TOKEN = "topological"


class PaginateFunction(Protocol):
    async def __call__(
        self,
        *,
        room_id: str,
        from_key: RoomStreamToken,
        to_key: RoomStreamToken | None = None,
        direction: Direction = Direction.BACKWARDS,
        limit: int = 0,
    ) -> tuple[list[EventBase], RoomStreamToken, bool]: ...


# Used as return values for pagination APIs
@attr.s(slots=True, frozen=True, auto_attribs=True)
class _EventDictReturn:
    event_id: str
    topological_ordering: int | None
    stream_ordering: int


@attr.s(slots=True, frozen=True, auto_attribs=True)
class _EventsAround:
    events_before: list[EventBase]
    events_after: list[EventBase]
    start: RoomStreamToken
    end: RoomStreamToken


@attr.s(slots=True, frozen=True, auto_attribs=True)
class CurrentStateDeltaMembership:
    """
    Attributes:
        event_id: The "current" membership event ID in this room.
        event_pos: The position of the "current" membership event in the event stream.
        prev_event_id: The previous membership event in this room that was replaced by
            the "current" one. May be `None` if there was no previous membership event.
        room_id: The room ID of the membership event.
        membership: The membership state of the user in the room
        sender: The person who sent the membership event
    """

    room_id: str
    # Event
    event_id: str | None
    event_pos: PersistedEventPosition
    membership: str
    sender: str | None
    # Prev event
    prev_event_id: str | None
    prev_event_pos: PersistedEventPosition | None
    prev_membership: str | None
    prev_sender: str | None


def generate_pagination_where_clause(
    direction: Direction,
    column_names: tuple[str, str],
    from_token: tuple[int | None, int] | None,
    to_token: tuple[int | None, int] | None,
    engine: BaseDatabaseEngine,
) -> str:
    """Creates an SQL expression to bound the columns by the pagination
    tokens.

    For example creates an SQL expression like:

        (6, 7) >= (topological_ordering, stream_ordering)
        AND (5, 3) < (topological_ordering, stream_ordering)

    would be generated for dir=b, from_token=(6, 7) and to_token=(5, 3).

    Note that tokens are considered to be after the row they are in, e.g. if
    a row A has a token T, then we consider A to be before T. This convention
    is important when figuring out inequalities for the generated SQL, and
    produces the following result:
        - If paginating forwards then we exclude any rows matching the from
          token, but include those that match the to token.
        - If paginating backwards then we include any rows matching the from
          token, but include those that match the to token.

    Args:
        direction: Whether we're paginating backwards or forwards.
        column_names: The column names to bound. Must *not* be user defined as
            these get inserted directly into the SQL statement without escapes.
        from_token: The start point for the pagination. This is an exclusive
            minimum bound if direction is forwards, and an inclusive maximum bound if
            direction is backwards.
        to_token: The endpoint point for the pagination. This is an inclusive
            maximum bound if direction is forwards, and an exclusive minimum bound if
            direction is backwards.
        engine: The database engine to generate the clauses for

    Returns:
        The sql expression
    """

    where_clause = []
    if from_token:
        where_clause.append(
            _make_generic_sql_bound(
                bound=">=" if direction == Direction.BACKWARDS else "<",
                column_names=column_names,
                values=from_token,
                engine=engine,
            )
        )

    if to_token:
        where_clause.append(
            _make_generic_sql_bound(
                bound="<" if direction == Direction.BACKWARDS else ">=",
                column_names=column_names,
                values=to_token,
                engine=engine,
            )
        )

    return " AND ".join(where_clause)


def generate_pagination_bounds(
    direction: Direction,
    from_token: RoomStreamToken | None,
    to_token: RoomStreamToken | None,
) -> tuple[str, tuple[int | None, int] | None, tuple[int | None, int] | None]:
    """
    Generate a start and end point for this page of events.

    Args:
        direction: Whether pagination is going forwards or backwards.
        from_token: The token to start pagination at, or None to start at the first value.
        to_token: The token to end pagination at, or None to not limit the end point.

    Returns:
        A three tuple of:

            ASC or DESC for sorting of the query.

            The starting position as a tuple of ints representing
            (topological position, stream position) or None if no from_token was
            provided. The topological position may be None for live tokens.

            The end position in the same format as the starting position, or None
            if no to_token was provided.
    """

    # Tokens really represent positions between elements, but we use
    # the convention of pointing to the event before the gap. Hence
    # we have a bit of asymmetry when it comes to equalities.
    if direction == Direction.BACKWARDS:
        order = "DESC"
    else:
        order = "ASC"

    # The bounds for the stream tokens are complicated by the fact
    # that we need to handle the instance_map part of the tokens. We do this
    # by fetching all events between the min stream token and the maximum
    # stream token (as returned by `RoomStreamToken.get_max_stream_pos`) and
    # then filtering the results.
    from_bound: tuple[int | None, int] | None = None
    if from_token:
        if from_token.topological is not None:
            from_bound = from_token.as_historical_tuple()
        elif direction == Direction.BACKWARDS:
            from_bound = (
                None,
                from_token.get_max_stream_pos(),
            )
        else:
            from_bound = (
                None,
                from_token.stream,
            )

    to_bound: tuple[int | None, int] | None = None
    if to_token:
        if to_token.topological is not None:
            to_bound = to_token.as_historical_tuple()
        elif direction == Direction.BACKWARDS:
            to_bound = (
                None,
                to_token.stream,
            )
        else:
            to_bound = (
                None,
                to_token.get_max_stream_pos(),
            )

    return order, from_bound, to_bound


def generate_next_token(
    direction: Direction, last_topo_ordering: int | None, last_stream_ordering: int
) -> RoomStreamToken:
    """
    Generate the next room stream token based on the currently returned data.

    Args:
        direction: Whether pagination is going forwards or backwards.
        last_topo_ordering: The last topological ordering being returned.
        last_stream_ordering: The last stream ordering being returned.

    Returns:
        A new RoomStreamToken to return to the client.
    """
    if direction == Direction.BACKWARDS:
        # Tokens are positions between events.
        # This token points *after* the last event in the chunk.
        # We need it to point to the event before it in the chunk
        # when we are going backwards so we subtract one from the
        # stream part.
        last_stream_ordering -= 1
    return RoomStreamToken(topological=last_topo_ordering, stream=last_stream_ordering)


def _make_generic_sql_bound(
    bound: str,
    column_names: tuple[str, str],
    values: tuple[int | None, int],
    engine: BaseDatabaseEngine,
) -> str:
    """Create an SQL expression that bounds the given column names by the
    values, e.g. create the equivalent of `(1, 2) < (col1, col2)`.

    Only works with two columns.

    Older versions of SQLite don't support that syntax so we have to expand it
    out manually.

    Args:
        bound: The comparison operator to use. One of ">", "<", ">=",
            "<=", where the values are on the left and columns on the right.
        names: The column names. Must *not* be user defined
            as these get inserted directly into the SQL statement without
            escapes.
        values: The values to bound the columns by. If
            the first value is None then only creates a bound on the second
            column.
        engine: The database engine to generate the SQL for

    Returns:
        The SQL statement
    """

    assert bound in (">", "<", ">=", "<=")

    name1, name2 = column_names
    val1, val2 = values

    if val1 is None:
        val2 = int(val2)
        return "(%d %s %s)" % (val2, bound, name2)

    val1 = int(val1)
    val2 = int(val2)

    if isinstance(engine, PostgresEngine):
        # Postgres doesn't optimise ``(x < a) OR (x=a AND y<b)`` as well
        # as it optimises ``(x,y) < (a,b)`` on multicolumn indexes. So we
        # use the later form when running against postgres.
        return "((%d,%d) %s (%s,%s))" % (val1, val2, bound, name1, name2)

    # We want to generate queries of e.g. the form:
    #
    #   (val1 < name1 OR (val1 = name1 AND val2 <= name2))
    #
    # which is equivalent to (val1, val2) < (name1, name2)

    return """(
        {val1:d} {strict_bound} {name1}
        OR ({val1:d} = {name1} AND {val2:d} {bound} {name2})
    )""".format(
        name1=name1,
        val1=val1,
        name2=name2,
        val2=val2,
        strict_bound=bound[0],  # The first bound must always be strict equality here
        bound=bound,
    )


def _filter_results(
    lower_token: RoomStreamToken | None,
    upper_token: RoomStreamToken | None,
    instance_name: str | None,
    topological_ordering: int,
    stream_ordering: int,
) -> bool:
    """Returns True if the event persisted by the given instance at the given
    topological/stream_ordering falls between the two tokens (taking a None
    token to mean unbounded).

    Used to filter results from fetching events in the DB against the given
    tokens. This is necessary to handle the case where the tokens include
    position maps, which we handle by fetching more than necessary from the DB
    and then filtering (rather than attempting to construct a complicated SQL
    query).

    The `instance_name` arg is optional to handle historic rows, and is
    interpreted as if it was "master".
    """

    if instance_name is None:
        instance_name = "master"

    event_historical_tuple = (
        topological_ordering,
        stream_ordering,
    )

    if lower_token:
        if lower_token.topological is not None:
            # If these are historical tokens we compare the `(topological, stream)`
            # tuples.
            if event_historical_tuple <= lower_token.as_historical_tuple():
                return False

        else:
            # If these are live tokens we compare the stream ordering against the
            # writers stream position.
            if stream_ordering <= lower_token.get_stream_pos_for_instance(
                instance_name
            ):
                return False

    if upper_token:
        if upper_token.topological is not None:
            if upper_token.as_historical_tuple() < event_historical_tuple:
                return False
        else:
            if upper_token.get_stream_pos_for_instance(instance_name) < stream_ordering:
                return False

    return True


def _filter_results_by_stream(
    lower_token: RoomStreamToken | None,
    upper_token: RoomStreamToken | None,
    instance_name: str | None,
    stream_ordering: int,
) -> bool:
    """
    This function only works with "live" tokens with `stream_ordering` only. See
    `_filter_results(...)` if you want to work with all tokens.

    Returns True if the event persisted by the given instance at the given
    stream_ordering falls between the two tokens (taking a None
    token to mean unbounded).

    The token range is defined by > `lower_token` and <= `upper_token`.

    Used to filter results from fetching events in the DB against the given
    tokens. This is necessary to handle the case where the tokens include
    position maps, which we handle by fetching more than necessary from the DB
    and then filtering (rather than attempting to construct a complicated SQL
    query).

    The `instance_name` arg is optional to handle historic rows, and is
    interpreted as if it was "master".
    """
    if instance_name is None:
        instance_name = "master"

    if lower_token:
        assert lower_token.topological is None

        # If these are live tokens we compare the stream ordering against the
        # writers stream position.
        if stream_ordering <= lower_token.get_stream_pos_for_instance(instance_name):
            return False

    if upper_token:
        assert upper_token.topological is None

        if upper_token.get_stream_pos_for_instance(instance_name) < stream_ordering:
            return False

    return True


def filter_to_clause(event_filter: Filter | None) -> tuple[str, list[str]]:
    # NB: This may create SQL clauses that don't optimise well (and we don't
    # have indices on all possible clauses). E.g. it may create
    # "room_id == X AND room_id != X", which postgres doesn't optimise.

    if not event_filter:
        return "", []

    clauses = []
    args = []

    # Handle event types with potential wildcard characters
    if event_filter.types:
        type_clauses = []
        for typ in event_filter.types:
            if "*" in typ:
                type_clauses.append("event.type LIKE ?")
                typ = typ.replace("*", "%")  # Replace * with % for SQL LIKE pattern
            else:
                type_clauses.append("event.type = ?")
            args.append(typ)
        clauses.append("(%s)" % " OR ".join(type_clauses))

    # Handle event types to exclude with potential wildcard characters
    for typ in event_filter.not_types:
        if "*" in typ:
            clauses.append("event.type NOT LIKE ?")
            typ = typ.replace("*", "%")
        else:
            clauses.append("event.type != ?")
        args.append(typ)

    if event_filter.senders:
        clauses.append(
            "(%s)" % " OR ".join("event.sender = ?" for _ in event_filter.senders)
        )
        args.extend(event_filter.senders)

    for sender in event_filter.not_senders:
        clauses.append("event.sender != ?")
        args.append(sender)

    if event_filter.rooms:
        clauses.append(
            "(%s)" % " OR ".join("event.room_id = ?" for _ in event_filter.rooms)
        )
        args.extend(event_filter.rooms)

    for room_id in event_filter.not_rooms:
        clauses.append("event.room_id != ?")
        args.append(room_id)

    if event_filter.contains_url:
        clauses.append("event.contains_url = ?")
        args.append(event_filter.contains_url)

    # We're only applying the "labels" filter on the database query, because applying the
    # "not_labels" filter via a SQL query is non-trivial. Instead, we let
    # event_filter.check_fields apply it, which is not as efficient but makes the
    # implementation simpler.
    if event_filter.labels:
        clauses.append("(%s)" % " OR ".join("label = ?" for _ in event_filter.labels))
        args.extend(event_filter.labels)

    # Filter on relation_senders / relation types from the joined tables.
    if event_filter.related_by_senders:
        clauses.append(
            "(%s)"
            % " OR ".join(
                "related_event.sender = ?" for _ in event_filter.related_by_senders
            )
        )
        args.extend(event_filter.related_by_senders)

    if event_filter.related_by_rel_types:
        clauses.append(
            "(%s)"
            % " OR ".join(
                "relation_type = ?" for _ in event_filter.related_by_rel_types
            )
        )
        args.extend(event_filter.related_by_rel_types)

    if event_filter.rel_types:
        clauses.append(
            "(%s)"
            % " OR ".join(
                "event_relation.relation_type = ?" for _ in event_filter.rel_types
            )
        )
        args.extend(event_filter.rel_types)

    if event_filter.not_rel_types:
        clauses.append(
            "((%s) OR event_relation.relation_type IS NULL)"
            % " AND ".join(
                "event_relation.relation_type != ?" for _ in event_filter.not_rel_types
            )
        )
        args.extend(event_filter.not_rel_types)

    return " AND ".join(clauses), args


class StreamWorkerStore(EventsWorkerStore, SQLBaseStore):
    def __init__(
        self,
        database: DatabasePool,
        db_conn: LoggingDatabaseConnection,
        hs: "HomeServer",
    ):
        super().__init__(database, db_conn, hs)

        self._instance_name = hs.get_instance_name()
        self._send_federation = hs.should_send_federation()
        self._federation_shard_config = hs.config.worker.federation_shard_config

        # If we're a process that sends federation we may need to reset the
        # `federation_stream_position` table to match the current sharding
        # config. We don't do this now as otherwise two processes could conflict
        # during startup which would cause one to die.
        self._need_to_reset_federation_stream_positions = self._send_federation

        events_max = self.get_room_max_stream_ordering()
        event_cache_prefill, min_event_val = self.db_pool.get_cache_dict(
            db_conn,
            "events",
            entity_column="room_id",
            stream_column="stream_ordering",
            max_value=events_max,
        )
        self._events_stream_cache = StreamChangeCache(
            name="EventsRoomStreamChangeCache",
            server_name=self.server_name,
            current_stream_pos=min_event_val,
            prefilled_cache=event_cache_prefill,
        )
        self._membership_stream_cache = StreamChangeCache(
            name="MembershipStreamChangeCache",
            server_name=self.server_name,
            current_stream_pos=events_max,
        )

        self._stream_order_on_start = self.get_room_max_stream_ordering()
        self._min_stream_order_on_start = self.get_room_min_stream_ordering()

    def get_room_max_stream_ordering(self) -> int:
        """Get the stream_ordering of regular events that we have committed up to

        Returns the maximum stream id such that all stream ids less than or
        equal to it have been successfully persisted.
        """
        return self._stream_id_gen.get_current_token()

    def get_room_min_stream_ordering(self) -> int:
        """Get the stream_ordering of backfilled events that we have committed up to

        Backfilled events use *negative* stream orderings, so this returns the
        minimum negative stream id such that all stream ids greater than or
        equal to it have been successfully persisted.
        """
        return self._backfill_id_gen.get_current_token()

    def get_room_max_token(self) -> RoomStreamToken:
        """Get a `RoomStreamToken` that marks the current maximum persisted
        position of the events stream. Useful to get a token that represents
        "now".

        The token returned is a "live" token that may have an instance_map
        component.
        """

        return RoomStreamToken.from_generator(self._stream_id_gen)

    def get_events_stream_id_generator(self) -> MultiWriterIdGenerator:
        return self._stream_id_gen

    async def get_room_events_stream_for_rooms(
        self,
        *,
        room_ids: Collection[str],
        from_key: RoomStreamToken,
        to_key: RoomStreamToken | None = None,
        direction: Direction = Direction.BACKWARDS,
        limit: int = 0,
    ) -> dict[str, tuple[list[EventBase], RoomStreamToken, bool]]:
        """Get new room events in stream ordering since `from_key`.

        Args:
            room_ids
            from_key: The token to stream from (starting point and heading in the given
                direction)
            to_key: The token representing the end stream position (end point)
            limit: Maximum number of events to return
            direction: Indicates whether we are paginating forwards or backwards
                from `from_key`.

        Returns:
            A map from room id to a tuple containing:
                - list of recent events in the room
                - stream ordering key for the start of the chunk of events returned.
                - a boolean to indicate if there were more events but we hit the limit

            When Direction.FORWARDS: from_key < x <= to_key, (ascending order)
            When Direction.BACKWARDS: from_key >= x > to_key, (descending order)
        """
        if direction == Direction.FORWARDS:
            room_ids = self._events_stream_cache.get_entities_changed(
                room_ids, from_key.stream
            )
        elif direction == Direction.BACKWARDS:
            if to_key is not None:
                room_ids = self._events_stream_cache.get_entities_changed(
                    room_ids, to_key.stream
                )
        else:
            assert_never(direction)

        if not room_ids:
            return {}

        results = {}
        room_ids = list(room_ids)
        for rm_ids in (room_ids[i : i + 20] for i in range(0, len(room_ids), 20)):
            res = await make_deferred_yieldable(
                defer.gatherResults(
                    [
                        run_in_background(
                            self.paginate_room_events_by_stream_ordering,
                            room_id=room_id,
                            from_key=from_key,
                            to_key=to_key,
                            direction=direction,
                            limit=limit,
                        )
                        for room_id in rm_ids
                    ],
                    consumeErrors=True,
                )
            )
            results.update(dict(zip(rm_ids, res)))

        return results

    def get_rooms_that_changed(
        self, room_ids: Collection[str], from_key: RoomStreamToken
    ) -> set[str]:
        """Given a list of rooms and a token, return rooms where there may have
        been changes.
        """
        from_id = from_key.stream
        return {
            room_id
            for room_id in room_ids
            if self._events_stream_cache.has_entity_changed(room_id, from_id)
        }

    async def get_rooms_that_have_updates_since_sliding_sync_table(
        self,
        room_ids: StrCollection,
        from_key: RoomStreamToken,
    ) -> StrCollection:
        """Return the rooms that probably have had updates since the given
        token (changes that are > `from_key`).

        May return false positives, but must not return false negatives.

        If `have_finished_sliding_sync_background_jobs` is False, then we return
        all the room IDs, as we can't be sure that the sliding sync table is
        fully populated.
        """
        # If the stream change cache is valid for the stream token, we can just
        # use the result of that.
        if from_key.stream >= self._events_stream_cache.get_earliest_known_position():
            return self._events_stream_cache.get_entities_changed(
                room_ids, from_key.stream
            )

        if not self.have_finished_sliding_sync_background_jobs():
            # If the table hasn't been populated yet, we have to assume all rooms
            # have updates.
            return room_ids

        def get_rooms_that_have_updates_since_sliding_sync_table_txn(
            txn: LoggingTransaction,
        ) -> StrCollection:
            sql = """
                SELECT room_id
                FROM sliding_sync_joined_rooms
                WHERE {clause}
                    AND event_stream_ordering > ?
            """

            results: set[str] = set()
            for batch in batch_iter(room_ids, 1000):
                clause, args = make_in_list_sql_clause(
                    self.database_engine, "room_id", batch
                )

                args.append(from_key.stream)
                txn.execute(sql.format(clause=clause), args)

                results.update(row[0] for row in txn)

            return results

        return await self.db_pool.runInteraction(
            "get_rooms_that_have_updates_since_sliding_sync_table",
            get_rooms_that_have_updates_since_sliding_sync_table_txn,
        )

    async def paginate_room_events_by_stream_ordering(
        self,
        *,
        room_id: str,
        from_key: RoomStreamToken,
        to_key: RoomStreamToken | None = None,
        direction: Direction = Direction.BACKWARDS,
        limit: int = 0,
    ) -> tuple[list[EventBase], RoomStreamToken, bool]:
        """
        Paginate events by `stream_ordering` in the room from the `from_key` in the
        given `direction` to the `to_key` or `limit`.

        Args:
            room_id
            from_key: The token to stream from (starting point and heading in the given
                direction)
            to_key: The token representing the end stream position (end point)
            direction: Indicates whether we are paginating forwards or backwards
                from `from_key`.
            limit: Maximum number of events to return

        Returns:
            The results as a list of events, a token that points to the end of
            the result set, and a boolean to indicate if there were more events
            but we hit the limit. If no events are returned then the end of the
            stream has been reached (i.e. there are no events between `from_key`
            and `to_key`).

            When Direction.FORWARDS: from_key < x <= to_key, (ascending order)
            When Direction.BACKWARDS: from_key >= x > to_key, (descending order)
        """

        # FIXME: When going forwards, we should enforce that the `to_key` is not `None`
        # because we always need an upper bound when querying the events stream (as
        # otherwise we'll potentially pick up events that are not fully persisted).

        # We should only be working with `stream_ordering` tokens here
        assert from_key is None or from_key.topological is None
        assert to_key is None or to_key.topological is None

        # We can bail early if we're looking forwards, and our `to_key` is already
        # before our `from_key`.
        if (
            direction == Direction.FORWARDS
            and to_key is not None
            and to_key.is_before_or_eq(from_key)
        ):
            # Token selection matches what we do below if there are no rows
            return [], to_key if to_key else from_key, False
        # Or vice-versa, if we're looking backwards and our `from_key` is already before
        # our `to_key`.
        elif (
            direction == Direction.BACKWARDS
            and to_key is not None
            and from_key.is_before_or_eq(to_key)
        ):
            # Token selection matches what we do below if there are no rows
            return [], to_key if to_key else from_key, False

        # We can do a quick sanity check to see if any events have been sent in the room
        # since the earlier token.
        has_changed = True
        if direction == Direction.FORWARDS:
            has_changed = self._events_stream_cache.has_entity_changed(
                room_id, from_key.stream
            )
        elif direction == Direction.BACKWARDS:
            if to_key is not None:
                has_changed = self._events_stream_cache.has_entity_changed(
                    room_id, to_key.stream
                )
        else:
            assert_never(direction)

        if not has_changed:
            # Token selection matches what we do below if there are no rows
            return [], to_key if to_key else from_key, False

        order, from_bound, to_bound = generate_pagination_bounds(
            direction, from_key, to_key
        )

        bounds = generate_pagination_where_clause(
            direction=direction,
            # The empty string will shortcut downstream code to only use the
            # `stream_ordering` column
            column_names=("", "stream_ordering"),
            from_token=from_bound,
            to_token=to_bound,
            engine=self.database_engine,
        )

        def f(txn: LoggingTransaction) -> tuple[list[_EventDictReturn], bool]:
            sql = f"""
                SELECT event_id, instance_name, stream_ordering
                FROM events
                WHERE
                    room_id = ?
                    AND not outlier
                    AND {bounds}
                ORDER BY stream_ordering {order} LIMIT ?
            """
            txn.execute(sql, (room_id, 2 * limit))

            # Get all the rows and check if we hit the limit.
            fetched_rows = txn.fetchall()
            limited = len(fetched_rows) >= 2 * limit

            rows = [
                _EventDictReturn(event_id, None, stream_ordering)
                for event_id, instance_name, stream_ordering in fetched_rows
                if _filter_results_by_stream(
                    lower_token=(
                        to_key if direction == Direction.BACKWARDS else from_key
                    ),
                    upper_token=(
                        from_key if direction == Direction.BACKWARDS else to_key
                    ),
                    instance_name=instance_name,
                    stream_ordering=stream_ordering,
                )
            ]

            if len(rows) > limit:
                limited = True

            rows = rows[:limit]
            return rows, limited

        rows, limited = await self.db_pool.runInteraction(
            "get_room_events_stream_for_room", f
        )

        ret = await self.get_events_as_list(
            [r.event_id for r in rows], get_prev_content=True
        )

        if rows:
            next_key = generate_next_token(
                direction=direction,
                last_topo_ordering=None,
                last_stream_ordering=rows[-1].stream_ordering,
            )
        else:
            # TODO (erikj): We should work out what to do here instead. (same as
            # `_paginate_room_events_by_topological_ordering_txn(...)`)
            next_key = to_key if to_key else from_key

        return ret, next_key, limited

    @trace
    async def get_current_state_delta_membership_changes_for_user(
        self,
        user_id: str,
        from_key: RoomStreamToken,
        to_key: RoomStreamToken,
        excluded_room_ids: list[str] | None = None,
    ) -> list[CurrentStateDeltaMembership]:
        """
        Fetch membership events (and the previous event that was replaced by that one)
        for a given user.

        Note: This function only works with "live" tokens with `stream_ordering` only.

        We're looking for membership changes in the token range (> `from_key` and <=
        `to_key`).

        Please be mindful to only use this with `from_key` and `to_key` tokens that are
        recent enough to be after when the first local user joined the room. Otherwise,
        the results may be incomplete or too greedy. For example, if you use a token
        range before the first local user joined the room, you will see 0 events since
        `current_state_delta_stream` tracks what the server thinks is the current state
        of the room as time goes. It does not track how state progresses from the
        beginning of the room. So for example, when you remotely join a room, the first
        rows will just be the state when you joined and progress from there.

        You can probably reasonably use this with `/sync` because the `to_key` passed in
        will be the "current" now token and the range will cover when the user joined
        the room.

        Args:
            user_id: The user ID to fetch membership events for.
            from_key: The point in the stream to sync from (fetching events > this point).
            to_key: The token to fetch rooms up to (fetching events <= this point).
            excluded_room_ids: Optional list of room IDs to exclude from the results.

        Returns:
            All membership changes to the current state in the token range. Events are
            sorted by `stream_ordering` ascending.

            `event_id`/`sender` can be `None` when the server leaves a room (meaning
            everyone locally left) or a state reset which removed the person from the
            room. We can't tell the difference between the two cases with what's
            available in the `current_state_delta_stream` table. To actually check for a
            state reset, you need to check if a membership still exists in the room.
        """

        assert from_key.topological is None
        assert to_key.topological is None

        # Start by ruling out cases where a DB query is not necessary.
        if from_key == to_key:
            return []

        if from_key:
            has_changed = self._membership_stream_cache.has_entity_changed(
                user_id, int(from_key.stream)
            )
            if not has_changed:
                return []

        def f(txn: LoggingTransaction) -> list[CurrentStateDeltaMembership]:
            # To handle tokens with a non-empty instance_map we fetch more
            # results than necessary and then filter down
            min_from_id = from_key.stream
            max_to_id = to_key.get_max_stream_pos()

            args: list[Any] = [min_from_id, max_to_id, EventTypes.Member, user_id]

            # TODO: It would be good to assert that the `from_token`/`to_token` is >=
            # the first row in `current_state_delta_stream` for the rooms we're
            # interested in. Otherwise, we will end up with empty results and not know
            # it.

            # We could `COALESCE(e.stream_ordering, s.stream_id)` to get more accurate
            # stream positioning when available but given our usages, we can avoid the
            # complexity. Between two (valid) stream tokens, we will still get all of
            # the state changes. Since those events are persisted in a batch, valid
            # tokens will either be before or after the batch of events.
            #
            # `stream_ordering` from the `events` table is more accurate when available
            # since the `current_state_delta_stream` table only tracks that the current
            # state is at this stream position (not what stream position the state event
            # was added) and uses the *minimum* stream position for batches of events.
            sql = """
                SELECT
                    s.room_id,
                    e.event_id,
                    s.instance_name,
                    s.stream_id,
                    m.membership,
                    e.sender,
                    s.prev_event_id,
                    e_prev.instance_name AS prev_instance_name,
                    e_prev.stream_ordering AS prev_stream_ordering,
                    m_prev.membership AS prev_membership,
                    e_prev.sender AS prev_sender
                FROM current_state_delta_stream AS s
                    LEFT JOIN events AS e ON e.event_id = s.event_id
                    LEFT JOIN room_memberships AS m ON m.event_id = s.event_id
                    LEFT JOIN events AS e_prev ON e_prev.event_id = s.prev_event_id
                    LEFT JOIN room_memberships AS m_prev ON m_prev.event_id = s.prev_event_id
                WHERE s.stream_id > ? AND s.stream_id <= ?
                    AND s.type = ?
                    AND s.state_key = ?
                ORDER BY s.stream_id ASC
            """

            txn.execute(sql, args)

            membership_changes: list[CurrentStateDeltaMembership] = []
            for (
                room_id,
                event_id,
                instance_name,
                stream_ordering,
                membership,
                sender,
                prev_event_id,
                prev_instance_name,
                prev_stream_ordering,
                prev_membership,
                prev_sender,
            ) in txn:
                assert room_id is not None
                assert stream_ordering is not None

                if _filter_results_by_stream(
                    from_key,
                    to_key,
                    instance_name,
                    stream_ordering,
                ):
                    # When the server leaves a room, it will insert new rows into the
                    # `current_state_delta_stream` table with `event_id = null` for all
                    # current state. This means we might already have a row for the
                    # leave event and then another for the same leave where the
                    # `event_id=null` but the `prev_event_id` is pointing back at the
                    # earlier leave event. We don't want to report the leave, if we
                    # already have a leave event.
                    if event_id is None and prev_membership == Membership.LEAVE:
                        continue

                    membership_change = CurrentStateDeltaMembership(
                        room_id=room_id,
                        # Event
                        event_id=event_id,
                        event_pos=PersistedEventPosition(
                            # If instance_name is null we default to "master"
                            instance_name=instance_name or "master",
                            stream=stream_ordering,
                        ),
                        # When `s.event_id = null`, we won't be able to get respective
                        # `room_membership` but can assume the user has left the room
                        # because this only happens when the server leaves a room
                        # (meaning everyone locally left) or a state reset which removed
                        # the person from the room.
                        membership=(
                            membership if membership is not None else Membership.LEAVE
                        ),
                        # This will also be null for the same reasons if `s.event_id = null`
                        sender=sender,
                        # Prev event
                        prev_event_id=prev_event_id,
                        prev_event_pos=(
                            PersistedEventPosition(
                                # If instance_name is null we default to "master"
                                instance_name=prev_instance_name or "master",
                                stream=prev_stream_ordering,
                            )
                            if (prev_stream_ordering is not None)
                            else None
                        ),
                        prev_membership=prev_membership,
                        prev_sender=prev_sender,
                    )

                    membership_changes.append(membership_change)

            return membership_changes

        membership_changes = await self.db_pool.runInteraction(
            "get_current_state_delta_membership_changes_for_user", f
        )

        room_ids_to_exclude: AbstractSet[str] = set()
        if excluded_room_ids is not None:
            room_ids_to_exclude = set(excluded_room_ids)

        return [
            membership_change
            for membership_change in membership_changes
            if membership_change.room_id not in room_ids_to_exclude
        ]

    @trace
    async def get_sliding_sync_membership_changes(
        self,
        user_id: str,
        from_key: RoomStreamToken,
        to_key: RoomStreamToken,
        excluded_room_ids: AbstractSet[str] | None = None,
    ) -> dict[str, RoomsForUserStateReset]:
        """
        Fetch membership events that result in a meaningful membership change for a
        given user.

        A meaningful membership changes is one where the `membership` value actually
        changes. This means memberships changes from `join` to `join` (like a display
        name change) will be filtered out since they result in no meaningful change.

        Note: This function only works with "live" tokens with `stream_ordering` only.

        We're looking for membership changes in the token range (> `from_key` and <=
        `to_key`).

        Args:
            user_id: The user ID to fetch membership events for.
            from_key: The point in the stream to sync from (fetching events > this point).
            to_key: The token to fetch rooms up to (fetching events <= this point).
            excluded_room_ids: Optional list of room IDs to exclude from the results.

        Returns:
            All meaningful membership changes to the current state in the token range.
            Events are sorted by `stream_ordering` ascending.

            `event_id`/`sender` can be `None` when the server leaves a room (meaning
            everyone locally left) or a state reset which removed the person from the
            room. We can't tell the difference between the two cases with what's
            available in the `current_state_delta_stream` table. To actually check for a
            state reset, you need to check if a membership still exists in the room.
        """

        assert from_key.topological is None
        assert to_key.topological is None

        # Start by ruling out cases where a DB query is not necessary.
        if from_key == to_key:
            return {}

        if from_key:
            has_changed = self._membership_stream_cache.has_entity_changed(
                user_id, int(from_key.stream)
            )
            if not has_changed:
                return {}

        room_ids_to_exclude: AbstractSet[str] = set()
        if excluded_room_ids is not None:
            room_ids_to_exclude = excluded_room_ids

        def f(txn: LoggingTransaction) -> dict[str, RoomsForUserStateReset]:
            # To handle tokens with a non-empty instance_map we fetch more
            # results than necessary and then filter down
            min_from_id = from_key.stream
            max_to_id = to_key.get_max_stream_pos()

            # This query looks at membership changes in
            # `sliding_sync_membership_snapshots` which will not include users
            # that were state reset out of rooms; so we need to look for that
            # case in `current_state_delta_stream`.
            sql = """
                SELECT
                    room_id,
                    membership_event_id,
                    event_instance_name,
                    event_stream_ordering,
                    membership,
                    sender,
                    prev_membership,
                    room_version
                FROM
                (
                    SELECT
                        s.room_id,
                        s.membership_event_id,
                        s.event_instance_name,
                        s.event_stream_ordering,
                        s.membership,
                        s.sender,
                        m_prev.membership AS prev_membership
                    FROM sliding_sync_membership_snapshots as s
                        LEFT JOIN event_edges AS e ON e.event_id = s.membership_event_id
                        LEFT JOIN room_memberships AS m_prev ON m_prev.event_id = e.prev_event_id
                    WHERE s.user_id = ?

                    UNION ALL

                    SELECT
                        s.room_id,
                        e.event_id,
                        s.instance_name,
                        s.stream_id,
                        m.membership,
                        e.sender,
                        m_prev.membership AS prev_membership
                    FROM current_state_delta_stream AS s
                        LEFT JOIN events AS e ON e.event_id = s.event_id
                        LEFT JOIN room_memberships AS m ON m.event_id = s.event_id
                        LEFT JOIN room_memberships AS m_prev ON m_prev.event_id = s.prev_event_id
                    WHERE
                        s.type = ?
                        AND s.state_key = ?
                ) AS c
                INNER JOIN rooms USING (room_id)
                WHERE event_stream_ordering > ? AND event_stream_ordering <= ?
                ORDER BY event_stream_ordering ASC
            """

            txn.execute(
                sql,
                (user_id, EventTypes.Member, user_id, min_from_id, max_to_id),
            )

            membership_changes: dict[str, RoomsForUserStateReset] = {}
            for (
                room_id,
                membership_event_id,
                event_instance_name,
                event_stream_ordering,
                membership,
                sender,
                prev_membership,
                room_version_id,
            ) in txn:
                assert room_id is not None
                assert event_stream_ordering is not None

                if room_id in room_ids_to_exclude:
                    continue

                if _filter_results_by_stream(
                    from_key,
                    to_key,
                    event_instance_name,
                    event_stream_ordering,
                ):
                    # When the server leaves a room, it will insert new rows into the
                    # `current_state_delta_stream` table with `event_id = null` for all
                    # current state. This means we might already have a row for the
                    # leave event and then another for the same leave where the
                    # `event_id=null` but the `prev_event_id` is pointing back at the
                    # earlier leave event. We don't want to report the leave, if we
                    # already have a leave event.
                    if (
                        membership_event_id is None
                        and prev_membership == Membership.LEAVE
                    ):
                        continue

                    if membership_event_id is None and room_id in membership_changes:
                        # SUSPICIOUS: if we join a room and get state reset out of it
                        # in the same queried window,
                        # won't this ignore the 'state reset out of it' part?
                        continue

                    # When `s.event_id = null`, we won't be able to get respective
                    # `room_membership` but can assume the user has left the room
                    # because this only happens when the server leaves a room
                    # (meaning everyone locally left) or a state reset which removed
                    # the person from the room.
                    membership = (
                        membership if membership is not None else Membership.LEAVE
                    )

                    if membership == prev_membership:
                        # If `membership` and `prev_membership` are the same then this
                        # is not a meaningful change so we can skip it.
                        # An example of this happening is when the user changes their display name.
                        continue

                    membership_change = RoomsForUserStateReset(
                        room_id=room_id,
                        sender=sender,
                        membership=membership,
                        event_id=membership_event_id,
                        event_pos=PersistedEventPosition(
                            event_instance_name, event_stream_ordering
                        ),
                        room_version_id=room_version_id,
                    )

                    membership_changes[room_id] = membership_change

            return membership_changes

        membership_changes = await self.db_pool.runInteraction(
            "get_sliding_sync_membership_changes", f
        )

        return membership_changes

    @cancellable
    async def get_membership_changes_for_user(
        self,
        user_id: str,
        from_key: RoomStreamToken,
        to_key: RoomStreamToken,
        excluded_rooms: list[str] | None = None,
    ) -> list[EventBase]:
        """Fetch membership events for a given user.

        All such events whose stream ordering `s` lies in the range
        `from_key < s <= to_key` are returned. Events are ordered by ascending stream
        order.
        """
        # Start by ruling out cases where a DB query is not necessary.
        if from_key == to_key:
            return []

        if from_key:
            has_changed = self._membership_stream_cache.has_entity_changed(
                user_id, int(from_key.stream)
            )
            if not has_changed:
                return []

        def f(txn: LoggingTransaction) -> list[_EventDictReturn]:
            # To handle tokens with a non-empty instance_map we fetch more
            # results than necessary and then filter down
            min_from_id = from_key.stream
            max_to_id = to_key.get_max_stream_pos()

            args: list[Any] = [user_id, min_from_id, max_to_id]

            ignore_room_clause = ""
            if excluded_rooms is not None and len(excluded_rooms) > 0:
                ignore_room_clause, ignore_room_args = make_in_list_sql_clause(
                    txn.database_engine, "e.room_id", excluded_rooms, negative=True
                )
                ignore_room_clause = f"AND {ignore_room_clause}"
                args += ignore_room_args

            sql = """
                SELECT m.event_id, instance_name, topological_ordering, stream_ordering
                FROM events AS e, room_memberships AS m
                WHERE e.event_id = m.event_id
                    AND m.user_id = ?
                    AND e.stream_ordering > ? AND e.stream_ordering <= ?
                    %s
                ORDER BY e.stream_ordering ASC
            """ % (ignore_room_clause,)

            txn.execute(sql, args)

            rows = [
                _EventDictReturn(event_id, None, stream_ordering)
                for event_id, instance_name, topological_ordering, stream_ordering in txn
                if _filter_results(
                    from_key,
                    to_key,
                    instance_name,
                    topological_ordering,
                    stream_ordering,
                )
            ]

            return rows

        rows = await self.db_pool.runInteraction("get_membership_changes_for_user", f)

        ret = await self.get_events_as_list(
            [r.event_id for r in rows], get_prev_content=True
        )

        return ret

    async def get_recent_events_for_room(
        self, room_id: str, limit: int, end_token: RoomStreamToken
    ) -> tuple[list[EventBase], RoomStreamToken]:
        """Get the most recent events in the room in topological ordering.

        Args:
            room_id
            limit
            end_token: The stream token representing now.

        Returns:
            A list of events and a token pointing to the start of the returned
            events. The events returned are in ascending topological order.
        """

        rows, token = await self.get_recent_event_ids_for_room(
            room_id, limit, end_token
        )

        events = await self.get_events_as_list(
            [r.event_id for r in rows], get_prev_content=True
        )

        return events, token

    async def get_recent_event_ids_for_room(
        self, room_id: str, limit: int, end_token: RoomStreamToken
    ) -> tuple[list[_EventDictReturn], RoomStreamToken]:
        """Get the most recent events in the room in topological ordering.

        Args:
            room_id
            limit
            end_token: The stream token representing now.

        Returns:
            A list of _EventDictReturn and a token pointing to the start of the
            returned events. The events returned are in ascending order.
        """
        # Allow a zero limit here, and no-op.
        if limit == 0:
            return [], end_token

        rows, token, _ = await self.db_pool.runInteraction(
            "get_recent_event_ids_for_room",
            self._paginate_room_events_by_topological_ordering_txn,
            room_id,
            from_token=end_token,
            limit=limit,
        )

        # We want to return the results in ascending order.
        rows.reverse()

        return rows, token

    async def get_room_event_before_stream_ordering(
        self, room_id: str, stream_ordering: int
    ) -> tuple[int, int, str] | None:
        """Gets details of the first event in a room at or before a stream ordering

        Args:
            room_id:
            stream_ordering:

        Returns:
            A tuple of (stream ordering, topological ordering, event_id)
        """

        def _f(txn: LoggingTransaction) -> tuple[int, int, str] | None:
            sql = """
                SELECT stream_ordering, topological_ordering, event_id
                FROM events
                LEFT JOIN rejections USING (event_id)
                WHERE room_id = ?
                    AND stream_ordering <= ?
                    AND NOT outlier
                    AND rejections.event_id IS NULL
                ORDER BY stream_ordering DESC
                LIMIT 1
            """
            txn.execute(sql, (room_id, stream_ordering))
            return cast(tuple[int, int, str] | None, txn.fetchone())

        return await self.db_pool.runInteraction(
            "get_room_event_before_stream_ordering", _f
        )

    async def get_last_event_id_in_room_before_stream_ordering(
        self,
        room_id: str,
        end_token: RoomStreamToken,
    ) -> str | None:
        """Returns the ID of the last event in a room at or before a stream ordering

        Args:
            room_id
            end_token: The token used to stream from

        Returns:
            The ID of the most recent event, or None if there are no events in the room
            before this stream ordering.
        """
        last_event_result = (
            await self.get_last_event_pos_in_room_before_stream_ordering(
                room_id, end_token
            )
        )

        if last_event_result:
            return last_event_result[0]

        return None

    async def get_last_event_pos_in_room(
        self,
        room_id: str,
        event_types: StrCollection | None = None,
    ) -> tuple[str, PersistedEventPosition] | None:
        """
        Returns the ID and event position of the last event in a room.

        Based on `get_last_event_pos_in_room_before_stream_ordering(...)`

        Args:
            room_id
            event_types: Optional allowlist of event types to filter by

        Returns:
            The ID of the most recent event and it's position, or None if there are no
            events in the room that match the given event types.
        """

        def _get_last_event_pos_in_room_txn(
            txn: LoggingTransaction,
        ) -> tuple[str, PersistedEventPosition] | None:
            event_type_clause = ""
            event_type_args: list[str] = []
            if event_types is not None and len(event_types) > 0:
                event_type_clause, event_type_args = make_in_list_sql_clause(
                    txn.database_engine, "type", event_types
                )
                event_type_clause = f"AND {event_type_clause}"

            sql = f"""
            SELECT event_id, stream_ordering, instance_name
            FROM events
            LEFT JOIN rejections USING (event_id)
            WHERE room_id = ?
                {event_type_clause}
                AND NOT outlier
                AND rejections.event_id IS NULL
            ORDER BY stream_ordering DESC
            LIMIT 1
            """

            txn.execute(
                sql,
                [room_id] + event_type_args,
            )

            row = cast(tuple[str, int, str] | None, txn.fetchone())
            if row is not None:
                event_id, stream_ordering, instance_name = row

                return event_id, PersistedEventPosition(
                    # If instance_name is null we default to "master"
                    instance_name or "master",
                    stream_ordering,
                )

            return None

        return await self.db_pool.runInteraction(
            "get_last_event_pos_in_room",
            _get_last_event_pos_in_room_txn,
        )

    @trace
    async def get_last_event_pos_in_room_before_stream_ordering(
        self,
        room_id: str,
        end_token: RoomStreamToken,
        event_types: StrCollection | None = None,
    ) -> tuple[str, PersistedEventPosition] | None:
        """
        Returns the ID and event position of the last event in a room at or before a
        stream ordering.

        Args:
            room_id
            end_token: The token used to stream from
            event_types: Optional allowlist of event types to filter by

        Returns:
            The ID of the most recent event and it's position, or None if there are no
            events in the room before this stream ordering.
        """

        def get_last_event_pos_in_room_before_stream_ordering_txn(
            txn: LoggingTransaction,
        ) -> tuple[str, PersistedEventPosition] | None:
            # We're looking for the closest event at or before the token. We need to
            # handle the fact that the stream token can be a vector clock (with an
            # `instance_map`) and events can be persisted on different instances
            # (sharded event persisters). The first subquery handles the events that
            # would be within the vector clock and gets all rows between the minimum and
            # maximum stream ordering in the token which need to be filtered against the
            # `instance_map`. The second subquery handles the "before" case and finds
            # the first row before the token. We then filter out any results past the
            # token's vector clock and return the first row that matches.
            min_stream = end_token.stream
            max_stream = end_token.get_max_stream_pos()

            event_type_clause = ""
            event_type_args: list[str] = []
            if event_types is not None and len(event_types) > 0:
                event_type_clause, event_type_args = make_in_list_sql_clause(
                    txn.database_engine, "type", event_types
                )
                event_type_clause = f"AND {event_type_clause}"

            # We use `UNION ALL` because we don't need any of the deduplication logic
            # (`UNION` is really a `UNION` + `DISTINCT`). `UNION ALL` does preserve the
            # ordering of the operand queries but there is no actual guarantee that it
            # has this behavior in all scenarios so we need the extra `ORDER BY` at the
            # bottom.
            sql = """
                SELECT * FROM (
                    SELECT instance_name, stream_ordering, topological_ordering, event_id
                    FROM events
                    LEFT JOIN rejections USING (event_id)
                    WHERE room_id = ?
                        %s
                        AND ? < stream_ordering AND stream_ordering <= ?
                        AND NOT outlier
                        AND rejections.event_id IS NULL
                    ORDER BY stream_ordering DESC
                ) AS a
                UNION ALL
                SELECT * FROM (
                    SELECT instance_name, stream_ordering, topological_ordering, event_id
                    FROM events
                    LEFT JOIN rejections USING (event_id)
                    WHERE room_id = ?
                        %s
                        AND stream_ordering <= ?
                        AND NOT outlier
                        AND rejections.event_id IS NULL
                    ORDER BY stream_ordering DESC
                    LIMIT 1
                ) AS b
                ORDER BY stream_ordering DESC
            """ % (
                event_type_clause,
                event_type_clause,
            )
            txn.execute(
                sql,
                [room_id]
                + event_type_args
                + [min_stream, max_stream, room_id]
                + event_type_args
                + [min_stream],
            )

            for instance_name, stream_ordering, topological_ordering, event_id in txn:
                if _filter_results(
                    lower_token=None,
                    upper_token=end_token,
                    instance_name=instance_name,
                    topological_ordering=topological_ordering,
                    stream_ordering=stream_ordering,
                ):
                    return event_id, PersistedEventPosition(
                        # If instance_name is null we default to "master"
                        instance_name or "master",
                        stream_ordering,
                    )

            return None

        return await self.db_pool.runInteraction(
            "get_last_event_pos_in_room_before_stream_ordering",
            get_last_event_pos_in_room_before_stream_ordering_txn,
        )

    async def bulk_get_last_event_pos_in_room_before_stream_ordering(
        self,
        room_ids: StrCollection,
        end_token: RoomStreamToken,
    ) -> dict[str, int]:
        """Bulk fetch the stream position of the latest events in the given
        rooms
        """

        # First we just get the latest positions for the room, as the vast
        # majority of them will be before the given end token anyway. By doing
        # this we can cache most rooms.
        uncapped_results = await self._bulk_get_max_event_pos(room_ids)

        # Check that the stream position for the rooms are from before the
        # minimum position of the token. If not then we need to fetch more
        # rows.
        results: dict[str, int] = {}
        recheck_rooms: set[str] = set()
        min_token = end_token.stream
        for room_id, stream in uncapped_results.items():
            if stream is None:
                # Despite the function not directly setting None, the cache can!
                # See: https://github.com/element-hq/synapse/issues/17726
                continue
            if stream <= min_token:
                results[room_id] = stream
            else:
                recheck_rooms.add(room_id)

        if not recheck_rooms:
            return results

        # There shouldn't be many rooms that we need to recheck, so we do them
        # one-by-one.
        for room_id in recheck_rooms:
            result = await self.get_last_event_pos_in_room_before_stream_ordering(
                room_id, end_token
            )
            if result is not None:
                results[room_id] = result[1].stream

        return results

    @cached()
    async def _get_max_event_pos(self, room_id: str) -> int:
        raise NotImplementedError()

    @cachedList(cached_method_name="_get_max_event_pos", list_name="room_ids")
    async def _bulk_get_max_event_pos(
        self, room_ids: StrCollection
    ) -> Mapping[str, int | None]:
        """Fetch the max position of a persisted event in the room."""

        # We need to be careful not to return positions ahead of the current
        # positions, so we get the current token now and cap our queries to it.
        now_token = self.get_room_max_token()
        max_pos = now_token.get_max_stream_pos()

        results: dict[str, int] = {}

        # First, we check for the rooms in the stream change cache to see if we
        # can just use the latest position from it.
        missing_room_ids: set[str] = set()
        for room_id in room_ids:
            stream_pos = self._events_stream_cache.get_max_pos_of_last_change(room_id)
            if stream_pos is not None:
                results[room_id] = stream_pos
            else:
                missing_room_ids.add(room_id)

        if not missing_room_ids:
            return results

        # Next, we query the stream position from the DB. At first we fetch all
        # positions less than the *max* stream pos in the token, then filter
        # them down. We do this as a) this is a cheaper query, and b) the vast
        # majority of rooms will have a latest token from before the min stream
        # pos.

        def bulk_get_max_event_pos_fallback_txn(
            txn: LoggingTransaction, batched_room_ids: StrCollection
        ) -> dict[str, int]:
            clause, args = make_in_list_sql_clause(
                self.database_engine, "room_id", batched_room_ids
            )
            sql = f"""
                SELECT room_id, (
                    SELECT stream_ordering FROM events AS e
                    LEFT JOIN rejections USING (event_id)
                    WHERE e.room_id = r.room_id
                        AND e.stream_ordering <= ?
                        AND NOT outlier
                        AND rejection_reason IS NULL
                    ORDER BY stream_ordering DESC
                    LIMIT 1
                )
                FROM rooms AS r
                WHERE {clause}
            """
            txn.execute(sql, [max_pos] + args)
            return {row[0]: row[1] for row in txn}

        # It's easier to look at the `sliding_sync_joined_rooms` table and avoid all of
        # the joins and sub-queries.
        def bulk_get_max_event_pos_from_sliding_sync_tables_txn(
            txn: LoggingTransaction, batched_room_ids: StrCollection
        ) -> dict[str, int]:
            clause, args = make_in_list_sql_clause(
                self.database_engine, "room_id", batched_room_ids
            )
            sql = f"""
                SELECT room_id, event_stream_ordering
                FROM sliding_sync_joined_rooms
                WHERE {clause}
                ORDER BY event_stream_ordering DESC
            """
            txn.execute(sql, args)
            return {row[0]: row[1] for row in txn}

        recheck_rooms: set[str] = set()
        for batched in batch_iter(room_ids, 1000):
            if await self.have_finished_sliding_sync_background_jobs():
                batch_results = await self.db_pool.runInteraction(
                    "bulk_get_max_event_pos_from_sliding_sync_tables_txn",
                    bulk_get_max_event_pos_from_sliding_sync_tables_txn,
                    batched,
                )
            else:
                batch_results = await self.db_pool.runInteraction(
                    "bulk_get_max_event_pos_fallback_txn",
                    bulk_get_max_event_pos_fallback_txn,
                    batched,
                )
            for room_id, stream_ordering in batch_results.items():
                if stream_ordering <= now_token.stream:
                    results[room_id] = stream_ordering
                else:
                    recheck_rooms.add(room_id)

        # We now need to handle rooms where the above query returned a stream
        # position that was potentially too new. This should happen very rarely
        # so we just query the rooms one-by-one
        for room_id in recheck_rooms:
            result = await self.get_last_event_pos_in_room_before_stream_ordering(
                room_id, now_token
            )
            if result is not None:
                results[room_id] = result[1].stream

        return results

    async def get_current_room_stream_token_for_room_id(
        self, room_id: str
    ) -> RoomStreamToken:
        """Returns the current position of the rooms stream (historic token)."""
        stream_ordering = self.get_room_max_stream_ordering()
        topo = await self.db_pool.runInteraction(
            "_get_max_topological_txn", self._get_max_topological_txn, room_id
        )
        return RoomStreamToken(topological=topo, stream=stream_ordering)

    @overload
    def get_stream_id_for_event_txn(
        self,
        txn: LoggingTransaction,
        event_id: str,
        allow_none: Literal[False] = False,
    ) -> int: ...

    @overload
    def get_stream_id_for_event_txn(
        self,
        txn: LoggingTransaction,
        event_id: str,
        allow_none: bool = False,
    ) -> int | None: ...

    def get_stream_id_for_event_txn(
        self,
        txn: LoggingTransaction,
        event_id: str,
        allow_none: bool = False,
    ) -> int | None:
        # Type ignore: we pass keyvalues a Dict[str, str]; the function wants
        # Dict[str, Any]. I think mypy is unhappy because Dict is invariant?
        return self.db_pool.simple_select_one_onecol_txn(  # type: ignore[call-overload]
            txn=txn,
            table="events",
            keyvalues={"event_id": event_id},
            retcol="stream_ordering",
            allow_none=allow_none,
        )

    async def get_position_for_event(self, event_id: str) -> PersistedEventPosition:
        """Get the persisted position for an event"""
        row = await self.db_pool.simple_select_one(
            table="events",
            keyvalues={"event_id": event_id},
            retcols=("stream_ordering", "instance_name"),
            desc="get_position_for_event",
        )

        return PersistedEventPosition(row[1] or "master", row[0])

    async def get_topological_token_for_event(self, event_id: str) -> RoomStreamToken:
        """The stream token for an event
        Args:
            event_id: The id of the event to look up a stream token for.
        Raises:
            StoreError if the event wasn't in the database.
        Returns:
            A `RoomStreamToken` topological token.
        """
        row = await self.db_pool.simple_select_one(
            table="events",
            keyvalues={"event_id": event_id},
            retcols=("stream_ordering", "topological_ordering"),
            desc="get_topological_token_for_event",
        )
        return RoomStreamToken(topological=row[1], stream=row[0])

    async def get_current_topological_token(self, room_id: str, stream_key: int) -> int:
        """Gets the topological token in a room after or at the given stream
        ordering.

        Args:
            room_id
            stream_key
        """
        if isinstance(self.database_engine, PostgresEngine):
            min_function = "LEAST"
        elif isinstance(self.database_engine, Sqlite3Engine):
            min_function = "MIN"
        else:
            raise RuntimeError(f"Unknown database engine {self.database_engine}")

        # This query used to be
        #    SELECT COALESCE(MIN(topological_ordering), 0) FROM events
        #    WHERE room_id = ? and events.stream_ordering >= {stream_key}
        # which returns 0 if the stream_key is newer than any event in
        # the room. That's not wrong, but it seems to interact oddly with backfill,
        # requiring a second call to /messages to actually backfill from a remote
        # homeserver.
        #
        # Instead, rollback the stream ordering to that after the most recent event in
        # this room.
        sql = f"""
            WITH fallback(max_stream_ordering) AS (
                SELECT MAX(stream_ordering)
                FROM events
                WHERE room_id = ?
            )
            SELECT COALESCE(MIN(topological_ordering), 0) FROM events
            WHERE
                room_id = ?
                AND events.stream_ordering >= {min_function}(
                    ?,
                    (SELECT max_stream_ordering FROM fallback)
                )
        """

        row = await self.db_pool.execute(
            "get_current_topological_token", sql, room_id, room_id, stream_key
        )
        return row[0][0] if row else 0

    def _get_max_topological_txn(self, txn: LoggingTransaction, room_id: str) -> int:
        txn.execute(
            "SELECT MAX(topological_ordering) FROM events WHERE room_id = ?",
            (room_id,),
        )

        rows = txn.fetchall()
        # An aggregate function like MAX() will always return one row per group
        # so we can safely rely on the lookup here. For example, when a we
        # lookup a `room_id` which does not exist, `rows` will look like
        # `[(None,)]`
        return rows[0][0] if rows[0][0] is not None else 0

    async def get_events_around(
        self,
        room_id: str,
        event_id: str,
        before_limit: int,
        after_limit: int,
        event_filter: Filter | None = None,
    ) -> _EventsAround:
        """Retrieve events and pagination tokens around a given event in a
        room.
        """

        results = await self.db_pool.runInteraction(
            "get_events_around",
            self._get_events_around_txn,
            room_id,
            event_id,
            before_limit,
            after_limit,
            event_filter,
        )

        events_before = await self.get_events_as_list(
            list(results["before"]["event_ids"]), get_prev_content=True
        )

        events_after = await self.get_events_as_list(
            list(results["after"]["event_ids"]), get_prev_content=True
        )

        return _EventsAround(
            events_before=events_before,
            events_after=events_after,
            start=results["before"]["token"],
            end=results["after"]["token"],
        )

    def _get_events_around_txn(
        self,
        txn: LoggingTransaction,
        room_id: str,
        event_id: str,
        before_limit: int,
        after_limit: int,
        event_filter: Filter | None,
    ) -> dict:
        """Retrieves event_ids and pagination tokens around a given event in a
        room.

        Args:
            room_id
            event_id
            before_limit
            after_limit
            event_filter

        Returns:
            dict
        """

        row = self.db_pool.simple_select_one_txn(
            txn,
            "events",
            keyvalues={"event_id": event_id, "room_id": room_id},
            retcols=("stream_ordering", "topological_ordering"),
        )
        stream_ordering = int(row[0])
        topological_ordering = int(row[1])

        # Paginating backwards includes the event at the token, but paginating
        # forward doesn't.
        before_token = RoomStreamToken(
            topological=topological_ordering - 1, stream=stream_ordering
        )

        after_token = RoomStreamToken(
            topological=topological_ordering, stream=stream_ordering
        )

        rows, start_token, _ = self._paginate_room_events_by_topological_ordering_txn(
            txn,
            room_id,
            before_token,
            direction=Direction.BACKWARDS,
            limit=before_limit,
            event_filter=event_filter,
        )
        events_before = [r.event_id for r in rows]

        rows, end_token, _ = self._paginate_room_events_by_topological_ordering_txn(
            txn,
            room_id,
            after_token,
            direction=Direction.FORWARDS,
            limit=after_limit,
            event_filter=event_filter,
        )
        events_after = [r.event_id for r in rows]

        return {
            "before": {"event_ids": events_before, "token": start_token},
            "after": {"event_ids": events_after, "token": end_token},
        }

    async def get_all_new_event_ids_stream(
        self,
        from_id: int,
        current_id: int,
        limit: int,
    ) -> tuple[int, dict[str, int | None]]:
        """Get all new events

        Returns all event ids with from_id < stream_ordering <= current_id.

        Args:
            from_id:  the stream_ordering of the last event we processed
            current_id:  the stream_ordering of the most recently processed event
            limit: the maximum number of events to return

        Returns:
            A tuple of (next_id, event_to_received_ts), where `next_id`
            is the next value to pass as `from_id` (it will either be the
            stream_ordering of the last returned event, or, if fewer than `limit`
            events were found, the `current_id`). The `event_to_received_ts` is
            a dictionary mapping event ID to the event `received_ts`, sorted by ascending
            stream_ordering.
        """

        def get_all_new_event_ids_stream_txn(
            txn: LoggingTransaction,
        ) -> tuple[int, dict[str, int | None]]:
            sql = (
                "SELECT e.stream_ordering, e.event_id, e.received_ts"
                " FROM events AS e"
                " WHERE"
                " ? < e.stream_ordering AND e.stream_ordering <= ?"
                " ORDER BY e.stream_ordering ASC"
                " LIMIT ?"
            )

            txn.execute(sql, (from_id, current_id, limit))
            rows = txn.fetchall()

            upper_bound = current_id
            if len(rows) == limit:
                upper_bound = rows[-1][0]

            event_to_received_ts: dict[str, int | None] = {
                row[1]: row[2] for row in rows
            }
            return upper_bound, event_to_received_ts

        upper_bound, event_to_received_ts = await self.db_pool.runInteraction(
            "get_all_new_event_ids_stream", get_all_new_event_ids_stream_txn
        )

        return upper_bound, event_to_received_ts

    async def get_federation_out_pos(self, typ: str) -> int:
        if self._need_to_reset_federation_stream_positions:
            await self.db_pool.runInteraction(
                "_reset_federation_positions_txn", self._reset_federation_positions_txn
            )
            self._need_to_reset_federation_stream_positions = False

        return await self.db_pool.simple_select_one_onecol(
            table="federation_stream_position",
            retcol="stream_id",
            keyvalues={"type": typ, "instance_name": self._instance_name},
            desc="get_federation_out_pos",
        )

    async def update_federation_out_pos(self, typ: str, stream_id: int) -> None:
        if self._need_to_reset_federation_stream_positions:
            await self.db_pool.runInteraction(
                "_reset_federation_positions_txn", self._reset_federation_positions_txn
            )
            self._need_to_reset_federation_stream_positions = False

        await self.db_pool.simple_update_one(
            table="federation_stream_position",
            keyvalues={"type": typ, "instance_name": self._instance_name},
            updatevalues={"stream_id": stream_id},
            desc="update_federation_out_pos",
        )

    def _reset_federation_positions_txn(self, txn: LoggingTransaction) -> None:
        """Fiddles with the `federation_stream_position` table to make it match
        the configured federation sender instances during start up.
        """

        # The federation sender instances may have changed, so we need to
        # massage the `federation_stream_position` table to have a row per type
        # per instance sending federation. If there is a mismatch we update the
        # table with the correct rows using the *minimum* stream ID seen. This
        # may result in resending of events/EDUs to remote servers, but that is
        # preferable to dropping them.

        if not self._send_federation:
            return

        # Pull out the configured instances. If we don't have a shard config then
        # we assume that we're the only instance sending.
        configured_instances = self._federation_shard_config.instances
        if not configured_instances:
            configured_instances = [self._instance_name]
        elif self._instance_name not in configured_instances:
            return

        instances_in_table = self.db_pool.simple_select_onecol_txn(
            txn,
            table="federation_stream_position",
            keyvalues={},
            retcol="instance_name",
        )

        if set(instances_in_table) == set(configured_instances):
            # Nothing to do
            return

        sql = """
            SELECT type, MIN(stream_id) FROM federation_stream_position
            GROUP BY type
        """
        txn.execute(sql)
        min_positions = dict(
            cast(Iterable[tuple[str, int]], txn)
        )  # Map from type -> min position

        # Ensure we do actually have some values here
        assert set(min_positions) == {"federation", "events"}

        sql = """
            DELETE FROM federation_stream_position
            WHERE NOT (%s)
        """
        clause, args = make_in_list_sql_clause(
            txn.database_engine, "instance_name", configured_instances
        )
        txn.execute(sql % (clause,), args)

        for typ, stream_id in min_positions.items():
            self.db_pool.simple_upsert_txn(
                txn,
                table="federation_stream_position",
                keyvalues={"type": typ, "instance_name": self._instance_name},
                values={"stream_id": stream_id},
            )

    def has_room_changed_since(self, room_id: str, stream_id: int) -> bool:
        return self._events_stream_cache.has_entity_changed(room_id, stream_id)

    def _paginate_room_events_by_topological_ordering_txn(
        self,
        txn: LoggingTransaction,
        room_id: str,
        from_token: RoomStreamToken,
        to_token: RoomStreamToken | None = None,
        direction: Direction = Direction.BACKWARDS,
        limit: int = 0,
        event_filter: Filter | None = None,
    ) -> tuple[list[_EventDictReturn], RoomStreamToken, bool]:
        """Returns list of events before or after a given token.

        Args:
            txn
            room_id
            from_token: The token used to stream from
            to_token: A token which if given limits the results to only those before
            direction: Indicates whether we are paginating forwards or backwards
                from `from_key`.
            limit: The maximum number of events to return.
            event_filter: If provided filters the events to
                those that match the filter.

        Returns:
            A list of _EventDictReturn, a token that points to the end of the
            result set, and a boolean to indicate if there were more events but
            we hit the limit.  If no events are returned then the end of the
            stream has been reached (i.e. there are no events between
            `from_token` and `to_token`), or `limit` is zero.
        """
        # We can bail early if we're looking forwards, and our `to_key` is already
        # before our `from_token`.
        if (
            direction == Direction.FORWARDS
            and to_token is not None
            and to_token.is_before_or_eq(from_token)
        ):
            # Token selection matches what we do below if there are no rows
            return [], to_token if to_token else from_token, False
        # Or vice-versa, if we're looking backwards and our `from_token` is already before
        # our `to_token`.
        elif (
            direction == Direction.BACKWARDS
            and to_token is not None
            and from_token.is_before_or_eq(to_token)
        ):
            # Token selection matches what we do below if there are no rows
            return [], to_token if to_token else from_token, False

        args: list[Any] = [room_id]

        order, from_bound, to_bound = generate_pagination_bounds(
            direction, from_token, to_token
        )

        bounds = generate_pagination_where_clause(
            direction=direction,
            column_names=("event.topological_ordering", "event.stream_ordering"),
            from_token=from_bound,
            to_token=to_bound,
            engine=self.database_engine,
        )

        filter_clause, filter_args = filter_to_clause(event_filter)

        if filter_clause:
            bounds += " AND " + filter_clause
            args.extend(filter_args)

        # We fetch more events as we'll filter the result set
        requested_limit = int(limit) * 2
        args.append(int(limit) * 2)

        select_keywords = "SELECT"
        join_clause = ""
        # Using DISTINCT in this SELECT query is quite expensive, because it
        # requires the engine to sort on the entire (not limited) result set,
        # i.e. the entire events table. Only use it in scenarios that could result
        # in the same event ID occurring multiple times in the results.
        needs_distinct = False
        if event_filter and event_filter.labels:
            # If we're not filtering on a label, then joining on event_labels will
            # return as many row for a single event as the number of labels it has. To
            # avoid this, only join if we're filtering on at least one label.
            join_clause += """
                LEFT JOIN event_labels
                USING (event_id, room_id, topological_ordering)
            """
            if len(event_filter.labels) > 1:
                # Multiple labels could cause the same event to appear multiple times.
                needs_distinct = True

        # If there is a relation_senders and relation_types filter join to the
        # relations table to get events related to the current event.
        if event_filter and (
            event_filter.related_by_senders or event_filter.related_by_rel_types
        ):
            # Filtering by relations could cause the same event to appear multiple
            # times (since there's no limit on the number of relations to an event).
            needs_distinct = True
            join_clause += """
                LEFT JOIN event_relations AS relation ON (event.event_id = relation.relates_to_id)
            """
            if event_filter.related_by_senders:
                join_clause += """
                    LEFT JOIN events AS related_event ON (relation.event_id = related_event.event_id)
                """

        # If there is a not_rel_types filter join to the relations table to get
        # the event's relation information.
        if event_filter and (event_filter.rel_types or event_filter.not_rel_types):
            join_clause += """
                LEFT JOIN event_relations AS event_relation USING (event_id)
            """

        if needs_distinct:
            select_keywords += " DISTINCT"

        sql = """
            %(select_keywords)s
                event.event_id, event.instance_name,
                event.topological_ordering, event.stream_ordering
            FROM events AS event
            %(join_clause)s
            WHERE event.outlier = FALSE AND event.room_id = ? AND %(bounds)s
            ORDER BY event.topological_ordering %(order)s,
            event.stream_ordering %(order)s LIMIT ?
        """ % {
            "select_keywords": select_keywords,
            "join_clause": join_clause,
            "bounds": bounds,
            "order": order,
        }
        txn.execute(sql, args)

        # Get all the rows and check if we hit the limit.
        fetched_rows = txn.fetchall()
        limited = len(fetched_rows) >= requested_limit

        # Filter the result set.
        rows = [
            _EventDictReturn(event_id, topological_ordering, stream_ordering)
            for event_id, instance_name, topological_ordering, stream_ordering in fetched_rows
            if _filter_results(
                lower_token=(
                    to_token if direction == Direction.BACKWARDS else from_token
                ),
                upper_token=(
                    from_token if direction == Direction.BACKWARDS else to_token
                ),
                instance_name=instance_name,
                topological_ordering=topological_ordering,
                stream_ordering=stream_ordering,
            )
        ]

        if len(rows) > limit:
            limited = True

        rows = rows[:limit]

        if rows:
            assert rows[-1].topological_ordering is not None
            next_token = generate_next_token(
                direction, rows[-1].topological_ordering, rows[-1].stream_ordering
            )
        else:
            # TODO (erikj): We should work out what to do here instead.
            next_token = to_token if to_token else from_token

        return rows, next_token, limited

    @trace
    @tag_args
    async def paginate_room_events_by_topological_ordering(
        self,
        *,
        room_id: str,
        from_key: RoomStreamToken,
        to_key: RoomStreamToken | None = None,
        direction: Direction = Direction.BACKWARDS,
        limit: int = 0,
        event_filter: Filter | None = None,
    ) -> tuple[list[EventBase], RoomStreamToken, bool]:
        """
        Paginate events by `topological_ordering` (tie-break with `stream_ordering`) in
        the room from the `from_key` in the given `direction` to the `to_key` or
        `limit`.

        Args:
            room_id
            from_key: The token to stream from (starting point and heading in the given
                direction)
            to_key: The token representing the end stream position (end point)
            direction: Indicates whether we are paginating forwards or backwards
                from `from_key`.
            limit: Maximum number of events to return
            event_filter: If provided filters the events to those that match the filter.

        Returns:
            - The results as a list of events;
            - a token that points to the end of the result set; and
            - a boolean to indicate if there were more events
              but we hit the limit (`limited`)

            If no events are returned and `limited` is false, then the end of the
            stream has been reached (i.e. there are no events between `from_key`
            and `to_key`).

            When `limited` is true, that means that more pagination can be attempted.
            Note that `limited` can be true even if no events are returned,
            because rejected events are filtered out after the limit check.

            When Direction.FORWARDS: from_key < x <= to_key, (ascending order)
            When Direction.BACKWARDS: from_key >= x > to_key, (descending order)
        """

        # FIXME: When going forwards, we should enforce that the `to_key` is not `None`
        # because we always need an upper bound when querying the events stream (as
        # otherwise we'll potentially pick up events that are not fully persisted).

        # We have these checks outside of the transaction function (txn) to save getting
        # a DB connection and switching threads if we don't need to.
        #
        # We can bail early if we're looking forwards, and our `to_key` is already
        # before our `from_key`.
        if (
            direction == Direction.FORWARDS
            and to_key is not None
            and to_key.is_before_or_eq(from_key)
        ):
            # Token selection matches what we do in `_paginate_room_events_txn` if there
            # are no rows
            return [], to_key if to_key else from_key, False
        # Or vice-versa, if we're looking backwards and our `from_key` is already before
        # our `to_key`.
        elif (
            direction == Direction.BACKWARDS
            and to_key is not None
            and from_key.is_before_or_eq(to_key)
        ):
            # Token selection matches what we do in `_paginate_room_events_txn` if there
            # are no rows
            return [], to_key if to_key else from_key, False

        rows, token, limited = await self.db_pool.runInteraction(
            "paginate_room_events_by_topological_ordering",
            self._paginate_room_events_by_topological_ordering_txn,
            room_id,
            from_key,
            to_key,
            direction,
            limit,
            event_filter,
        )

        events = await self.get_events_as_list(
            [r.event_id for r in rows], get_prev_content=True
        )

        return events, token, limited

    @cached()
    async def get_id_for_instance(self, instance_name: str) -> int:
        """Get a unique, immutable ID that corresponds to the given Synapse worker instance."""

        def _get_id_for_instance_txn(txn: LoggingTransaction) -> int:
            instance_id = self.db_pool.simple_select_one_onecol_txn(
                txn,
                table="instance_map",
                keyvalues={"instance_name": instance_name},
                retcol="instance_id",
                allow_none=True,
            )
            if instance_id is not None:
                return instance_id

            # If we don't have an entry upsert one.
            #
            # We could do this before the first check, and rely on the cache for
            # efficiency, but each UPSERT causes the next ID to increment which
            # can quickly bloat the size of the generated IDs for new instances.
            self.db_pool.simple_upsert_txn(
                txn,
                table="instance_map",
                keyvalues={"instance_name": instance_name},
                values={},
            )

            return self.db_pool.simple_select_one_onecol_txn(
                txn,
                table="instance_map",
                keyvalues={"instance_name": instance_name},
                retcol="instance_id",
            )

        return await self.db_pool.runInteraction(
            "get_id_for_instance", _get_id_for_instance_txn
        )

    @cached()
    async def get_name_from_instance_id(self, instance_id: int) -> str:
        """Get the instance name from an ID previously returned by
        `get_id_for_instance`.
        """

        return await self.db_pool.simple_select_one_onecol(
            table="instance_map",
            keyvalues={"instance_id": instance_id},
            retcol="instance_name",
            desc="get_name_from_instance_id",
        )

    async def get_timeline_gaps(
        self,
        room_id: str,
        from_token: RoomStreamToken | None,
        to_token: RoomStreamToken,
    ) -> RoomStreamToken | None:
        """Check if there is a gap, and return a token that marks the position
        of the gap in the stream.
        """

        sql = """
            SELECT instance_name, stream_ordering
            FROM timeline_gaps
            WHERE room_id = ? AND ? < stream_ordering AND stream_ordering <= ?
            ORDER BY stream_ordering
        """

        rows = await self.db_pool.execute(
            "get_timeline_gaps",
            sql,
            room_id,
            from_token.stream if from_token else 0,
            to_token.get_max_stream_pos(),
        )

        if not rows:
            return None

        positions = [
            PersistedEventPosition(instance_name, stream_ordering)
            for instance_name, stream_ordering in rows
        ]
        if from_token:
            positions = [p for p in positions if p.persisted_after(from_token)]

        positions = [p for p in positions if not p.persisted_after(to_token)]

        if positions:
            # We return a stream token that ensures the event *at* the position
            # of the gap is included (as the gap is *before* the persisted
            # event).
            last_position = positions[-1]
            return RoomStreamToken(stream=last_position.stream - 1)

        return None

    @trace
    def get_rooms_that_might_have_updates(
        self, room_ids: StrCollection, from_token: RoomStreamToken
    ) -> StrCollection:
        """Filters given room IDs down to those that might have updates, i.e.
        removes rooms that definitely do not have updates.
        """
        return self._events_stream_cache.get_entities_changed(
            room_ids, from_token.stream
        )



# Copyright 2023 The Matrix.org Foundation C.I.C.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
import typing
from typing import cast

from synapse.api.auth.mas import MasDelegatedAuth
from synapse.api.errors import Codes, SynapseError
from synapse.http.server import HttpServer
from synapse.http.servlet import RestServlet
from synapse.http.site import SynapseRequest
from synapse.rest.client._base import client_patterns
from synapse.types import JsonDict

if typing.TYPE_CHECKING:
    from synapse.server import HomeServer


logger = logging.getLogger(__name__)


class AuthIssuerServlet(RestServlet):
    """
    Advertises what OpenID Connect issuer clients should use to authorise users.
    This endpoint was defined in a previous iteration of MSC2965, and is still
    used by some clients.
    """

    PATTERNS = client_patterns(
        "/org.matrix.msc2965/auth_issuer$",
        unstable=True,
        releases=(),
    )

    def __init__(self, hs: "HomeServer"):
        super().__init__()
        self._config = hs.config
        self._auth = hs.get_auth()

    async def on_GET(self, request: SynapseRequest) -> tuple[int, JsonDict]:
        # This endpoint is unauthenticated and the response only depends on
        # the metadata we get from Matrix Authentication Service. Internally,
        # MasDelegatedAuth/MSC3861DelegatedAuth.issuer() are already caching the
        # response in memory anyway. Ideally we would follow any Cache-Control directive
        # given by MAS, but this is fine for now.
        #
        # - `public` means it can be cached both in the browser and in caching proxies
        # - `max-age` controls how long we cache on the browser side. 10m is sane enough
        # - `s-maxage` controls how long we cache on the proxy side. Since caching
        #   proxies usually have a way to purge caches, it is fine to cache there for
        #   longer (1h), and issue cache invalidations in case we need it
        # - `stale-while-revalidate` allows caching proxies to serve stale content while
        #   revalidating in the background. This is useful for making this request always
        #   'snappy' to end users whilst still keeping it fresh
        request.setHeader(
            b"Cache-Control",
            b"public, max-age=600, s-maxage=3600, stale-while-revalidate=600",
        )

        if self._config.mas.enabled:
            assert isinstance(self._auth, MasDelegatedAuth)
            return 200, {"issuer": await self._auth.issuer()}

        elif self._config.experimental.msc3861.enabled:
            # If MSC3861 is enabled, we can assume self._auth is an instance of MSC3861DelegatedAuth
            # We import lazily here because of the authlib requirement
            from synapse.api.auth.msc3861_delegated import MSC3861DelegatedAuth

            assert isinstance(self._auth, MSC3861DelegatedAuth)
            return 200, {"issuer": await self._auth.issuer()}

        else:
            # Wouldn't expect this to be reached: the servelet shouldn't have been
            # registered. Still, fail gracefully if we are registered for some reason.
            raise SynapseError(
                404,
                "OIDC discovery has not been configured on this homeserver",
                Codes.NOT_FOUND,
            )


class AuthMetadataServlet(RestServlet):
    """
    Advertises the OAuth 2.0 server metadata for the homeserver.
    """

    PATTERNS = [
        *client_patterns(
            "/auth_metadata$",
            releases=("v1",),
        ),
        *client_patterns(
            "/org.matrix.msc2965/auth_metadata$",
            unstable=True,
            releases=(),
        ),
    ]

    def __init__(self, hs: "HomeServer"):
        super().__init__()
        self._config = hs.config
        self._auth = hs.get_auth()

    async def on_GET(self, request: SynapseRequest) -> tuple[int, JsonDict]:
        # This endpoint is unauthenticated and the response only depends on
        # the metadata we get from Matrix Authentication Service. Internally,
        # MasDelegatedAuth/MSC3861DelegatedAuth.issuer() are already caching the
        # response in memory anyway. Ideally we would follow any Cache-Control directive
        # given by MAS, but this is fine for now.
        #
        # - `public` means it can be cached both in the browser and in caching proxies
        # - `max-age` controls how long we cache on the browser side. 10m is sane enough
        # - `s-maxage` controls how long we cache on the proxy side. Since caching
        #   proxies usually have a way to purge caches, it is fine to cache there for
        #   longer (1h), and issue cache invalidations in case we need it
        # - `stale-while-revalidate` allows caching proxies to serve stale content while
        #   revalidating in the background. This is useful for making this request always
        #   'snappy' to end users whilst still keeping it fresh
        request.setHeader(
            b"Cache-Control",
            b"public, max-age=600, s-maxage=3600, stale-while-revalidate=600",
        )

        if self._config.mas.enabled:
            assert isinstance(self._auth, MasDelegatedAuth)
            return 200, await self._auth.auth_metadata()

        elif self._config.experimental.msc3861.enabled:
            # If MSC3861 is enabled, we can assume self._auth is an instance of MSC3861DelegatedAuth
            # We import lazily here because of the authlib requirement
            from synapse.api.auth.msc3861_delegated import MSC3861DelegatedAuth

            auth = cast(MSC3861DelegatedAuth, self._auth)
            return 200, await auth.auth_metadata()

        else:
            # Wouldn't expect this to be reached: the servlet shouldn't have been
            # registered. Still, fail gracefully if we are registered for some reason.
            raise SynapseError(
                404,
                "OIDC discovery has not been configured on this homeserver",
                Codes.NOT_FOUND,
            )


def register_servlets(hs: "HomeServer", http_server: HttpServer) -> None:
    if hs.config.mas.enabled or hs.config.experimental.msc3861.enabled:
        AuthIssuerServlet(hs).register(http_server)
        AuthMetadataServlet(hs).register(http_server)



#
# This file is licensed under the Affero General Public License (AGPL) version 3.
#
# Copyright 2015, 2016 OpenMarket Ltd
# Copyright (C) 2023 New Vector, Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# See the GNU Affero General Public License for more details:
# <https://www.gnu.org/licenses/agpl-3.0.html>.
#
# Originally licensed under the Apache License, Version 2.0:
# <http://www.apache.org/licenses/LICENSE-2.0>.
#
# [This file includes modifications made by New Vector Limited]
#
#

import logging
from typing import TYPE_CHECKING

from synapse.api.errors import AuthError, NotFoundError, StoreError, SynapseError
from synapse.http.server import HttpServer
from synapse.http.servlet import RestServlet, parse_json_object_from_request
from synapse.http.site import SynapseRequest
from synapse.types import JsonDict, JsonMapping, UserID

from ._base import client_patterns, set_timeline_upper_limit

if TYPE_CHECKING:
    from synapse.server import HomeServer

logger = logging.getLogger(__name__)


class GetFilterRestServlet(RestServlet):
    PATTERNS = client_patterns("/user/(?P<user_id>[^/]*)/filter/(?P<filter_id>[^/]*)")
    CATEGORY = "Encryption requests"

    def __init__(self, hs: "HomeServer"):
        super().__init__()
        self.hs = hs
        self.auth = hs.get_auth()
        self.filtering = hs.get_filtering()

    async def on_GET(
        self, request: SynapseRequest, user_id: str, filter_id: str
    ) -> tuple[int, JsonMapping]:
        target_user = UserID.from_string(user_id)
        requester = await self.auth.get_user_by_req(request)

        if target_user != requester.user:
            raise AuthError(403, "Cannot get filters for other users")

        if not self.hs.is_mine(target_user):
            raise AuthError(403, "Can only get filters for local users")

        try:
            filter_id_int = int(filter_id)
        except Exception:
            raise SynapseError(400, "Invalid filter_id")

        try:
            filter_collection = await self.filtering.get_user_filter(
                user_id=target_user, filter_id=filter_id_int
            )
        except StoreError as e:
            if e.code != 404:
                raise
            raise NotFoundError("No such filter")

        return 200, filter_collection.get_filter_json()


class CreateFilterRestServlet(RestServlet):
    PATTERNS = client_patterns("/user/(?P<user_id>[^/]*)/filter")
    CATEGORY = "Encryption requests"

    def __init__(self, hs: "HomeServer"):
        super().__init__()
        self.hs = hs
        self.auth = hs.get_auth()
        self.filtering = hs.get_filtering()

    async def on_POST(
        self, request: SynapseRequest, user_id: str
    ) -> tuple[int, JsonDict]:
        target_user = UserID.from_string(user_id)
        requester = await self.auth.get_user_by_req(request)

        if target_user != requester.user:
            raise AuthError(403, "Cannot create filters for other users")

        if not self.hs.is_mine(target_user):
            raise AuthError(403, "Can only create filters for local users")

        content = parse_json_object_from_request(request)
        set_timeline_upper_limit(content, self.hs.config.server.filter_timeline_limit)

        filter_id = await self.filtering.add_user_filter(
            user_id=target_user, user_filter=content
        )

        return 200, {"filter_id": str(filter_id)}


def register_servlets(hs: "HomeServer", http_server: HttpServer) -> None:
    GetFilterRestServlet(hs).register(http_server)
    CreateFilterRestServlet(hs).register(http_server)



#
# This file is licensed under the Affero General Public License (AGPL) version 3.
#
# Copyright (C) 2023 New Vector, Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# See the GNU Affero General Public License for more details:
# <https://www.gnu.org/licenses/agpl-3.0.html>.
#
# Originally licensed under the Apache License, Version 2.0:
# <http://www.apache.org/licenses/LICENSE-2.0>.
#
# [This file includes modifications made by New Vector Limited]
#
#
"""
The Federation Sender is responsible for sending Persistent Data Units (PDUs)
and Ephemeral Data Units (EDUs) to other homeservers using
the `/send` Federation API.


## How do PDUs get sent?

The Federation Sender is made aware of new PDUs due to `FederationSender.notify_new_events`.
When the sender is notified about a newly-persisted PDU that originates from this homeserver
and is not an out-of-band event, we pass the PDU to the `_PerDestinationQueue` for each
remote homeserver that is in the room at that point in the DAG.


### Per-Destination Queues

There is one `PerDestinationQueue` per 'destination' homeserver.
The `PerDestinationQueue` maintains the following information about the destination:

- whether the destination is currently in [catch-up mode (see below)](#catch-up-mode);
- a queue of PDUs to be sent to the destination; and
- a queue of EDUs to be sent to the destination (not considered in this section).

Upon a new PDU being enqueued, `attempt_new_transaction` is called to start a new
transaction if there is not already one in progress.


### Transactions and the Transaction Transmission Loop

Each federation HTTP request to the `/send` endpoint is referred to as a 'transaction'.
The body of the HTTP request contains a list of PDUs and EDUs to send to the destination.

The *Transaction Transmission Loop* (`_transaction_transmission_loop`) is responsible
for emptying the queued PDUs (and EDUs) from a `PerDestinationQueue` by sending
them to the destination.

There can only be one transaction in flight for a given destination at any time.
(Other than preventing us from overloading the destination, this also makes it easier to
reason about because we process events sequentially for each destination.
This is useful for *Catch-Up Mode*, described later.)

The loop continues so long as there is anything to send. At each iteration of the loop, we:

- dequeue up to 50 PDUs (and up to 100 EDUs).
- make the `/send` request to the destination homeserver with the dequeued PDUs and EDUs.
- if successful, make note of the fact that we succeeded in transmitting PDUs up to
  the given `stream_ordering` of the latest PDU by
- if unsuccessful, back off from the remote homeserver for some time.
  If we have been unsuccessful for too long (when the backoff interval grows to exceed 1 hour),
  the in-memory queues are emptied and we enter [*Catch-Up Mode*, described below](#catch-up-mode).


### Catch-Up Mode

When the `PerDestinationQueue` has the catch-up flag set, the *Catch-Up Transmission Loop*
(`_catch_up_transmission_loop`) is used in lieu of the regular `_transaction_transmission_loop`.
(Only once the catch-up mode has been exited can the regular transaction transmission behaviour
be resumed.)

*Catch-Up Mode*, entered upon Synapse startup or once a homeserver has fallen behind due to
connection problems, is responsible for sending PDUs that have been missed by the destination
homeserver. (PDUs can be missed because the `PerDestinationQueue` is volatile — i.e. resets
on startup — and it does not hold PDUs forever if `/send` requests to the destination fail.)

The catch-up mechanism makes use of the `last_successful_stream_ordering` column in the
`destinations` table (which gives the `stream_ordering` of the most recent successfully
sent PDU) and the `stream_ordering` column in the `destination_rooms` table (which gives,
for each room, the `stream_ordering` of the most recent PDU that needs to be sent to this
destination).

Each iteration of the loop pulls out 50 `destination_rooms` entries with the oldest
`stream_ordering`s that are greater than the `last_successful_stream_ordering`.
In other words, from the set of latest PDUs in each room to be sent to the destination,
the 50 oldest such PDUs are pulled out.

These PDUs could, in principle, now be directly sent to the destination. However, as an
optimisation intended to prevent overloading destination homeservers, we instead attempt
to send the latest forward extremities so long as the destination homeserver is still
eligible to receive those.
This reduces load on the destination **in aggregate** because all Synapse homeservers
will behave according to this principle and therefore avoid sending lots of different PDUs
at different points in the DAG to a recovering homeserver.
*This optimisation is not currently valid in rooms which are partial-state on this homeserver,
since we are unable to determine whether the destination homeserver is eligible to receive
the latest forward extremities unless this homeserver sent those PDUs — in this case, we
just send the latest PDUs originating from this server and skip this optimisation.*

Whilst PDUs are sent through this mechanism, the position of `last_successful_stream_ordering`
is advanced as normal.
Once there are no longer any rooms containing outstanding PDUs to be sent to the destination
*that are not already in the `PerDestinationQueue` because they arrived since Catch-Up Mode
was enabled*, Catch-Up Mode is exited and we return to `_transaction_transmission_loop`.


#### A note on failures and back-offs

If a remote server is unreachable over federation, we back off from that server,
with an exponentially-increasing retry interval.
We automatically retry after the retry interval expires (roughly, the logic to do so
being triggered every minute).

If the backoff grows too large (> 1 hour), the in-memory queue is emptied (to prevent
unbounded growth) and Catch-Up Mode is entered.

It is worth noting that the back-off for a remote server is cleared once an inbound
request from that remote server is received (see `notify_remote_server_up`).
At this point, the transaction transmission loop is also started up, to proactively
send missed PDUs and EDUs to the destination (i.e. you don't need to wait for a new PDU
or EDU, destined for that destination, to be created in order to send out missed PDUs and
EDUs).
"""

import abc
import logging
from collections import OrderedDict
from typing import (
    TYPE_CHECKING,
    Collection,
    Hashable,
    Iterable,
    Literal,
)

import attr
from prometheus_client import Counter

from twisted.internet import defer

import synapse.metrics
from synapse.api.constants import EventTypes, Membership
from synapse.api.presence import UserPresenceState
from synapse.events import EventBase
from synapse.federation.sender.per_destination_queue import (
    CATCHUP_RETRY_INTERVAL,
    PerDestinationQueue,
)
from synapse.federation.sender.transaction_manager import TransactionManager
from synapse.federation.units import Edu
from synapse.logging.context import make_deferred_yieldable, run_in_background
from synapse.metrics import (
    SERVER_NAME_LABEL,
    LaterGauge,
    event_processing_loop_counter,
    event_processing_loop_room_count,
    events_processed_counter,
)
from synapse.metrics.background_process_metrics import (
    wrap_as_background_process,
)
from synapse.types import (
    JsonDict,
    ReadReceipt,
    RoomStreamToken,
    StrCollection,
    get_domain_from_id,
)
from synapse.util.clock import Clock
from synapse.util.duration import Duration
from synapse.util.metrics import Measure
from synapse.util.retryutils import filter_destinations_by_retry_limiter

if TYPE_CHECKING:
    from synapse.events.presence_router import PresenceRouter
    from synapse.server import HomeServer

logger = logging.getLogger(__name__)

sent_pdus_destination_dist_count = Counter(
    "synapse_federation_client_sent_pdu_destinations_count",
    "Number of PDUs queued for sending to one or more destinations",
    labelnames=[SERVER_NAME_LABEL],
)

sent_pdus_destination_dist_total = Counter(
    "synapse_federation_client_sent_pdu_destinations",
    "Total number of PDUs queued for sending across all destinations",
    labelnames=[SERVER_NAME_LABEL],
)

transaction_queue_pending_destinations_gauge = LaterGauge(
    name="synapse_federation_transaction_queue_pending_destinations",
    desc="",
    labelnames=[SERVER_NAME_LABEL],
)

transaction_queue_pending_pdus_gauge = LaterGauge(
    name="synapse_federation_transaction_queue_pending_pdus",
    desc="",
    labelnames=[SERVER_NAME_LABEL],
)

transaction_queue_pending_edus_gauge = LaterGauge(
    name="synapse_federation_transaction_queue_pending_edus",
    desc="",
    labelnames=[SERVER_NAME_LABEL],
)

# Time (in s) to wait before trying to wake up destinations that have
# catch-up outstanding.
# Please note that rate limiting still applies, so while the loop is
# executed every X seconds the destinations may not be woken up because
# they are being rate limited following previous attempt failures.
WAKEUP_RETRY_PERIOD = Duration(minutes=1)

# Time to wait in between waking up each destination, i.e. one destination
# will be woken up every <x> seconds until we have woken every destination
# has outstanding catch-up.
WAKEUP_INTERVAL_BETWEEN_DESTINATIONS = Duration(seconds=5)


class AbstractFederationSender(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def shutdown(self) -> None:
        """Stops this federation sender instance from sending further transactions."""
        raise NotImplementedError()

    @abc.abstractmethod
    def notify_new_events(self, max_token: RoomStreamToken) -> None:
        """This gets called when we have some new events we might want to
        send out to other servers.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    async def send_read_receipt(self, receipt: ReadReceipt) -> None:
        """Send a RR to any other servers in the room

        Args:
            receipt: receipt to be sent
        """
        raise NotImplementedError()

    @abc.abstractmethod
    async def send_presence_to_destinations(
        self, states: Iterable[UserPresenceState], destinations: Iterable[str]
    ) -> None:
        """Send the given presence states to the given destinations.

        Args:
            destinations:
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def build_and_send_edu(
        self,
        destination: str,
        edu_type: str,
        content: JsonDict,
        key: Hashable | None = None,
    ) -> None:
        """Construct an Edu object, and queue it for sending

        Args:
            destination: name of server to send to
            edu_type: type of EDU to send
            content: content of EDU
            key: clobbering key for this edu
        """
        raise NotImplementedError()

    @abc.abstractmethod
    async def send_device_messages(
        self, destinations: StrCollection, immediate: bool = True
    ) -> None:
        """Tells the sender that a new device message is ready to be sent to the
        destinations. The `immediate` flag specifies whether the messages should
        be tried to be sent immediately, or whether it can be delayed for a
        short while (to aid performance).
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def wake_destination(self, destination: str) -> None:
        """Called when we want to retry sending transactions to a remote.

        This is mainly useful if the remote server has been down and we think it
        might have come back.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def get_current_token(self) -> int:
        raise NotImplementedError()

    @abc.abstractmethod
    def federation_ack(self, instance_name: str, token: int) -> None:
        raise NotImplementedError()

    @abc.abstractmethod
    async def get_replication_rows(
        self, instance_name: str, from_token: int, to_token: int, target_row_count: int
    ) -> tuple[list[tuple[int, tuple]], int, bool]:
        raise NotImplementedError()


@attr.s
class _DestinationWakeupQueue:
    """A queue of destinations that need to be woken up due to new updates.

    Staggers waking up of per destination queues to ensure that we don't attempt
    to start TLS connections with many hosts all at once, leading to pinned CPU.

    """

    # The maximum duration in seconds between queuing up a destination and it
    # being woken up.
    _MAX_TIME_IN_QUEUE = 30.0

    sender: "FederationSender" = attr.ib()
    hs: "HomeServer" = attr.ib()
    server_name: str = attr.ib()
    """
    Our homeserver name (used to label metrics) (`hs.hostname`).
    """
    clock: Clock = attr.ib()
    max_delay_s: int = attr.ib()

    queue: "OrderedDict[str, Literal[None]]" = attr.ib(factory=OrderedDict)
    processing: bool = attr.ib(default=False)

    def add_to_queue(self, destination: str) -> None:
        """Add a destination to the queue to be woken up."""

        self.queue[destination] = None

        if not self.processing:
            self._handle()

    @wrap_as_background_process("_DestinationWakeupQueue.handle")
    async def _handle(self) -> None:
        """Background process to drain the queue."""

        if not self.queue:
            return

        assert not self.processing
        self.processing = True

        try:
            # We start with a delay that should drain the queue quickly enough that
            # we process all destinations in the queue in _MAX_TIME_IN_QUEUE
            # seconds.
            #
            # We also add an upper bound to the delay, to gracefully handle the
            # case where the queue only has a few entries in it.
            current_sleep_seconds = min(
                self.max_delay_s, self._MAX_TIME_IN_QUEUE / len(self.queue)
            )

            while self.queue:
                destination, _ = self.queue.popitem(last=False)

                queue = self.sender._get_per_destination_queue(destination)
                if queue is None:
                    continue

                if not queue._new_data_to_send:
                    # The per destination queue has already been woken up.
                    continue

                queue.attempt_new_transaction()

                await self.clock.sleep(Duration(seconds=current_sleep_seconds))

                if not self.queue:
                    break

                # More destinations may have been added to the queue, so we may
                # need to reduce the delay to ensure everything gets processed
                # within _MAX_TIME_IN_QUEUE seconds.
                current_sleep_seconds = min(
                    current_sleep_seconds, self._MAX_TIME_IN_QUEUE / len(self.queue)
                )

        finally:
            self.processing = False


class FederationSender(AbstractFederationSender):
    def __init__(self, hs: "HomeServer"):
        self.hs = hs
        self.server_name = hs.hostname

        self.store = hs.get_datastores().main
        self.state = hs.get_state_handler()

        self._storage_controllers = hs.get_storage_controllers()

        self.clock = hs.get_clock()
        self.is_mine_id = hs.is_mine_id
        self.is_mine_server_name = hs.is_mine_server_name

        self._presence_router: "PresenceRouter" | None = None
        self._transaction_manager = TransactionManager(hs)

        self._instance_name = hs.get_instance_name()
        self._federation_shard_config = hs.config.worker.federation_shard_config

        # map from destination to PerDestinationQueue
        self._per_destination_queues: dict[str, PerDestinationQueue] = {}

        transaction_queue_pending_destinations_gauge.register_hook(
            homeserver_instance_id=hs.get_instance_id(),
            hook=lambda: {
                (self.server_name,): sum(
                    1
                    for d in self._per_destination_queues.values()
                    if d.transmission_loop_running
                )
            },
        )
        transaction_queue_pending_pdus_gauge.register_hook(
            homeserver_instance_id=hs.get_instance_id(),
            hook=lambda: {
                (self.server_name,): sum(
                    d.pending_pdu_count() for d in self._per_destination_queues.values()
                )
            },
        )
        transaction_queue_pending_edus_gauge.register_hook(
            homeserver_instance_id=hs.get_instance_id(),
            hook=lambda: {
                (self.server_name,): sum(
                    d.pending_edu_count() for d in self._per_destination_queues.values()
                )
            },
        )

        self._is_processing = False
        self._last_poked_id = -1

        self._external_cache = hs.get_external_cache()

        rr_txn_interval_per_room_s = (
            1.0 / hs.config.ratelimiting.federation_rr_transactions_per_room_per_second
        )
        self._destination_wakeup_queue = _DestinationWakeupQueue(
            self,
            hs,
            self.server_name,
            self.clock,
            max_delay_s=rr_txn_interval_per_room_s,
        )

        # It is important for `_is_shutdown` to be instantiated before the looping call
        # for `wake_destinations_needing_catchup`.
        self._is_shutdown = False

        # Regularly wake up destinations that have outstanding PDUs to be caught up
        self.clock.looping_call_now(
            self.hs.run_as_background_process,
            WAKEUP_RETRY_PERIOD,
            "wake_destinations_needing_catchup",
            self._wake_destinations_needing_catchup,
        )

    def shutdown(self) -> None:
        self._is_shutdown = True
        for queue in self._per_destination_queues.values():
            queue.shutdown()

    def _get_per_destination_queue(
        self, destination: str
    ) -> PerDestinationQueue | None:
        """Get or create a PerDestinationQueue for the given destination

        Args:
            destination: server_name of remote server

        Returns:
            None if the destination is not allowed by the federation whitelist.
            Otherwise a PerDestinationQueue for this destination.
        """
        if not self.hs.config.federation.is_domain_allowed_according_to_federation_whitelist(
            destination
        ):
            return None

        queue = self._per_destination_queues.get(destination)
        if not queue:
            queue = PerDestinationQueue(self.hs, self._transaction_manager, destination)
            self._per_destination_queues[destination] = queue
        return queue

    def notify_new_events(self, max_token: RoomStreamToken) -> None:
        """This gets called when we have some new events we might want to
        send out to other servers.
        """
        # We just use the minimum stream ordering and ignore the vector clock
        # component. This is safe to do as long as we *always* ignore the vector
        # clock components.
        current_id = max_token.stream

        self._last_poked_id = max(current_id, self._last_poked_id)

        if self._is_processing:
            return

        # fire off a processing loop in the background
        self.hs.run_as_background_process(
            "process_event_queue_for_federation",
            self._process_event_queue_loop,
        )

    async def _process_event_queue_loop(self) -> None:
        try:
            self._is_processing = True
            while not self._is_shutdown:
                last_token = await self.store.get_federation_out_pos("events")
                (
                    next_token,
                    event_to_received_ts,
                ) = await self.store.get_all_new_event_ids_stream(
                    last_token, self._last_poked_id, limit=100
                )

                event_ids = event_to_received_ts.keys()
                event_entries = await self.store.get_unredacted_events_from_cache_or_db(
                    event_ids
                )

                logger.debug(
                    "Handling %i -> %i: %i events to send (current id %i)",
                    last_token,
                    next_token,
                    len(event_entries),
                    self._last_poked_id,
                )

                if not event_entries and next_token >= self._last_poked_id:
                    logger.debug("All events processed")
                    break

                async def handle_event(event: EventBase) -> None:
                    # Only send events for this server.
                    send_on_behalf_of = event.internal_metadata.get_send_on_behalf_of()
                    is_mine = self.is_mine_id(event.sender)
                    if not is_mine and send_on_behalf_of is None:
                        logger.debug("Not sending remote-origin event %s", event)
                        return

                    # We also want to not send out-of-band membership events.
                    #
                    # OOB memberships are used in three (and a half) situations:
                    #
                    # (1) invite events which we have received over federation. Those
                    #     will have a `sender` on a different server, so will be
                    #     skipped by the "is_mine" test above anyway.
                    #
                    # (2) rejections of invites to federated rooms - either remotely
                    #     or locally generated. (Such rejections are normally
                    #     created via federation, in which case the remote server is
                    #     responsible for sending out the rejection. If that fails,
                    #     we'll create a leave event locally, but that's only really
                    #     for the benefit of the invited user - we don't have enough
                    #     information to send it out over federation).
                    #
                    # (2a) rescinded knocks. These are identical to rejected invites.
                    #
                    # (3) knock events which we have sent over federation. As with
                    #     invite rejections, the remote server should send them out to
                    #     the federation.
                    #
                    # So, in all the above cases, we want to ignore such events.
                    #
                    # OOB memberships are always(?) outliers anyway, so if we *don't*
                    # ignore them, we'll get an exception further down when we try to
                    # fetch the membership list for the room.
                    #
                    # Arguably, we could equivalently ignore all outliers here, since
                    # in theory the only way for an outlier with a local `sender` to
                    # exist is by being an OOB membership (via one of (2), (2a) or (3)
                    # above).
                    #
                    if event.internal_metadata.is_out_of_band_membership():
                        logger.debug("Not sending OOB membership event %s", event)
                        return

                    # Finally, there are some other events that we should not send out
                    # until someone asks for them. They are explicitly flagged as such
                    # with `proactively_send: False`.
                    if not event.internal_metadata.should_proactively_send():
                        logger.debug(
                            "Not sending event with proactively_send=false: %s", event
                        )
                        return

                    destinations: Collection[str] | None = None
                    if not event.prev_event_ids():
                        # If there are no prev event IDs then the state is empty
                        # and so no remote servers in the room
                        destinations = set()

                    if destinations is None:
                        # During partial join we use the set of servers that we got
                        # when beginning the join. It's still possible that we send
                        # events to servers that left the room in the meantime, but
                        # we consider that an acceptable risk since it is only our own
                        # events that we leak and not other server's ones.
                        partial_state_destinations = (
                            await self.store.get_partial_state_servers_at_join(
                                event.room_id
                            )
                        )

                        if partial_state_destinations is not None:
                            destinations = partial_state_destinations

                    if destinations is None:
                        # We check the external cache for the destinations, which is
                        # stored per state group.

                        sg = await self._external_cache.get(
                            "event_to_prev_state_group", event.event_id
                        )
                        if sg:
                            destinations = await self._external_cache.get(
                                "get_joined_hosts", str(sg)
                            )
                            if destinations is None:
                                # Add logging to help track down https://github.com/matrix-org/synapse/issues/13444
                                logger.info(
                                    "Unexpectedly did not have cached destinations for %s / %s",
                                    sg,
                                    event.event_id,
                                )
                        else:
                            # Add logging to help track down https://github.com/matrix-org/synapse/issues/13444
                            logger.info(
                                "Unexpectedly did not have cached prev group for %s",
                                event.event_id,
                            )

                    if destinations is None:
                        try:
                            # Get the state from before the event.
                            # We need to make sure that this is the state from before
                            # the event and not from after it.
                            # Otherwise if the last member on a server in a room is
                            # banned then it won't receive the event because it won't
                            # be in the room after the ban.
                            destinations = await self.state.get_hosts_in_room_at_events(
                                event.room_id, event_ids=event.prev_event_ids()
                            )
                        except Exception:
                            logger.exception(
                                "Failed to calculate hosts in room for event: %s",
                                event.event_id,
                            )
                            return

                    # If we've rescinded an invite then we want to tell the
                    # other server.
                    if (
                        event.type == EventTypes.Member
                        and event.membership == Membership.LEAVE
                        and event.sender != event.state_key
                    ):
                        # We check if this leave event is rescinding an invite
                        # by looking if there is an invite event for the user in
                        # the auth events. It could otherwise be a kick or
                        # unban, which we don't want to send (if the user wasn't
                        # already in the room).
                        auth_events = await self.store.get_events_as_list(
                            event.auth_event_ids()
                        )
                        for auth_event in auth_events:
                            if (
                                auth_event.type == EventTypes.Member
                                and auth_event.state_key == event.state_key
                                and auth_event.membership == Membership.INVITE
                            ):
                                destinations = set(destinations)
                                destinations.add(get_domain_from_id(event.state_key))
                                break

                    sharded_destinations = {
                        d
                        for d in destinations
                        if self._federation_shard_config.should_handle(
                            self._instance_name, d
                        )
                    }

                    if send_on_behalf_of is not None:
                        # If we are sending the event on behalf of another server
                        # then it already has the event and there is no reason to
                        # send the event to it.
                        sharded_destinations.discard(send_on_behalf_of)

                    logger.debug("Sending %s to %r", event, sharded_destinations)

                    if sharded_destinations:
                        await self._send_pdu(event, sharded_destinations)

                        now = self.clock.time_msec()
                        ts = event_to_received_ts[event.event_id]
                        assert ts is not None
                        synapse.metrics.event_processing_lag_by_event.labels(
                            name="federation_sender",
                            **{SERVER_NAME_LABEL: self.server_name},
                        ).observe((now - ts) / 1000)

                async def handle_room_events(events: list[EventBase]) -> None:
                    logger.debug(
                        "Handling %i events in room %s", len(events), events[0].room_id
                    )
                    with Measure(
                        self.clock,
                        name="handle_room_events",
                        server_name=self.server_name,
                    ):
                        for event in events:
                            await handle_event(event)

                events_by_room: dict[str, list[EventBase]] = {}

                for event_id in event_ids:
                    # `event_entries` is unsorted, so we have to iterate over `event_ids`
                    # to ensure the events are in the right order
                    event_cache = event_entries.get(event_id)
                    if event_cache:
                        event = event_cache.event
                        events_by_room.setdefault(event.room_id, []).append(event)

                await make_deferred_yieldable(
                    defer.gatherResults(
                        [
                            run_in_background(handle_room_events, evs)
                            for evs in events_by_room.values()
                        ],
                        consumeErrors=True,
                    )
                )

                logger.debug("Successfully handled up to %i", next_token)
                await self.store.update_federation_out_pos("events", next_token)

                if event_entries:
                    now = self.clock.time_msec()
                    ts = max(t for t in event_to_received_ts.values() if t)
                    assert ts is not None

                    synapse.metrics.event_processing_lag.labels(
                        name="federation_sender",
                        **{SERVER_NAME_LABEL: self.server_name},
                    ).set(now - ts)
                    synapse.metrics.event_processing_last_ts.labels(
                        name="federation_sender",
                        **{SERVER_NAME_LABEL: self.server_name},
                    ).set(ts)

                    events_processed_counter.labels(
                        **{SERVER_NAME_LABEL: self.server_name}
                    ).inc(len(event_entries))

                    event_processing_loop_room_count.labels(
                        name="federation_sender",
                        **{SERVER_NAME_LABEL: self.server_name},
                    ).inc(len(events_by_room))

                event_processing_loop_counter.labels(
                    name="federation_sender",
                    **{SERVER_NAME_LABEL: self.server_name},
                ).inc()

                synapse.metrics.event_processing_positions.labels(
                    name="federation_sender", **{SERVER_NAME_LABEL: self.server_name}
                ).set(next_token)

        finally:
            self._is_processing = False

    async def _send_pdu(self, pdu: EventBase, destinations: Iterable[str]) -> None:
        # We loop through all destinations to see whether we already have
        # a transaction in progress. If we do, stick it in the pending_pdus
        # table and we'll get back to it later.

        destinations = set(destinations)
        destinations.discard(self.server_name)
        logger.debug("Sending to: %s", str(destinations))

        if not destinations:
            return

        sent_pdus_destination_dist_total.labels(
            **{SERVER_NAME_LABEL: self.server_name}
        ).inc(len(destinations))
        sent_pdus_destination_dist_count.labels(
            **{SERVER_NAME_LABEL: self.server_name}
        ).inc()

        assert pdu.internal_metadata.stream_ordering

        # track the fact that we have a PDU for these destinations,
        # to allow us to perform catch-up later on if the remote is unreachable
        # for a while.
        # Filter out any destinations not present in the federation_domain_whitelist, if
        # the whitelist exists. These destinations should not be sent to so let's not
        # waste time or space keeping track of events destined for them.
        destinations = [
            d
            for d in destinations
            if self.hs.config.federation.is_domain_allowed_according_to_federation_whitelist(
                d
            )
        ]
        await self.store.store_destination_rooms_entries(
            destinations,
            pdu.room_id,
            pdu.internal_metadata.stream_ordering,
        )

        destinations = await filter_destinations_by_retry_limiter(
            destinations,
            clock=self.clock,
            store=self.store,
            retry_due_within_ms=CATCHUP_RETRY_INTERVAL,
        )

        for destination in destinations:
            queue = self._get_per_destination_queue(destination)
            # We expect `queue` to not be None as we already filtered out
            # non-whitelisted destinations above.
            assert queue is not None

            queue.send_pdu(pdu)

    async def send_read_receipt(self, receipt: ReadReceipt) -> None:
        """Send a RR to any other servers in the room

        Args:
            receipt: receipt to be sent
        """

        # Some background on the rate-limiting going on here.
        #
        # It turns out that if we attempt to send out RRs as soon as we get them
        # from a client, then we end up trying to do several hundred Hz of
        # federation transactions. (The number of transactions scales as O(N^2)
        # on the size of a room, since in a large room we have both more RRs
        # coming in, and more servers to send them to.)
        #
        # This leads to a lot of CPU load, and we end up getting behind. The
        # solution currently adopted is to differentiate between receipts and
        # destinations we should immediately send to, and those we can trickle
        # the receipts to.
        #
        # The current logic is to send receipts out immediately if:
        #   - the room is "small", i.e. there's only N servers to send receipts
        #     to, and so sending out the receipts immediately doesn't cause too
        #     much load; or
        #   - the receipt is for an event that happened recently, as users
        #     notice if receipts are delayed when they know other users are
        #     currently reading the room; or
        #   - the receipt is being sent to the server that sent the event, so
        #     that users see receipts for their own receipts quickly.
        #
        # For destinations that we should delay sending the receipt to, we queue
        # the receipts up to be sent in the next transaction, but don't trigger
        # a new transaction to be sent. We then add the destination to the
        # `DestinationWakeupQueue`, which will slowly iterate over each
        # destination and trigger a new transaction to be sent.
        #
        # However, in practice, it is often possible to send out delayed
        # receipts earlier: in particular, if we are sending a transaction to a
        # given server anyway (for example, because we have a PDU or a RR in
        # another room to send), then we may as well send out all of the pending
        # RRs for that server. So it may be that by the time we get to waking up
        # the destination, we don't actually have any RRs left to send out.
        #
        # For even more background, see
        # https://github.com/matrix-org/synapse/issues/4730.

        room_id = receipt.room_id

        # Local read receipts always have 1 event ID.
        event_id = receipt.event_ids[0]

        # Work out which remote servers should be poked and poke them.
        domains_set = await self._storage_controllers.state.get_current_hosts_in_room_or_partial_state_approximation(
            room_id
        )
        domains: StrCollection = [
            d
            for d in domains_set
            if not self.is_mine_server_name(d)
            and self._federation_shard_config.should_handle(self._instance_name, d)
        ]

        domains = await filter_destinations_by_retry_limiter(
            domains,
            clock=self.clock,
            store=self.store,
            retry_due_within_ms=CATCHUP_RETRY_INTERVAL,
        )

        if not domains:
            return

        # We now split which domains we want to wake up immediately vs which we
        # want to delay waking up.
        immediate_domains: StrCollection
        delay_domains: StrCollection

        if len(domains) < 10:
            # For "small" rooms send to all domains immediately
            immediate_domains = domains
            delay_domains = ()
        else:
            metadata = await self.store.get_metadata_for_event(
                receipt.room_id, event_id
            )
            assert metadata is not None

            sender_domain = get_domain_from_id(metadata.sender)

            if self.clock.time_msec() - metadata.received_ts < 60_000:
                # We always send receipts for recent messages immediately
                immediate_domains = domains
                delay_domains = ()
            else:
                # Otherwise, we delay waking up all destinations except for the
                # sender's domain.
                immediate_domains = []
                delay_domains = []
                for domain in domains:
                    if domain == sender_domain:
                        immediate_domains.append(domain)
                    else:
                        delay_domains.append(domain)

        for domain in immediate_domains:
            # Add to destination queue and wake the destination up
            queue = self._get_per_destination_queue(domain)
            if queue is None:
                continue
            queue.queue_read_receipt(receipt)
            queue.attempt_new_transaction()

        for domain in delay_domains:
            # Add to destination queue...
            queue = self._get_per_destination_queue(domain)
            if queue is None:
                continue
            queue.queue_read_receipt(receipt)

            # ... and schedule the destination to be woken up.
            self._destination_wakeup_queue.add_to_queue(domain)

    async def send_presence_to_destinations(
        self, states: Iterable[UserPresenceState], destinations: Iterable[str]
    ) -> None:
        """Send the given presence states to the given destinations.
        destinations (list[str])
        """

        if not states or not self.hs.config.server.track_presence:
            # No-op if presence is disabled.
            return

        # Ensure we only send out presence states for local users.
        for state in states:
            assert self.is_mine_id(state.user_id)

        destinations = await filter_destinations_by_retry_limiter(
            [
                d
                for d in destinations
                if self._federation_shard_config.should_handle(self._instance_name, d)
            ],
            clock=self.clock,
            store=self.store,
            retry_due_within_ms=CATCHUP_RETRY_INTERVAL,
        )

        for destination in destinations:
            if self.is_mine_server_name(destination):
                continue

            queue = self._get_per_destination_queue(destination)
            if queue is None:
                continue
            queue.send_presence(states, start_loop=False)

            self._destination_wakeup_queue.add_to_queue(destination)

    def build_and_send_edu(
        self,
        destination: str,
        edu_type: str,
        content: JsonDict,
        key: Hashable | None = None,
    ) -> None:
        """Construct an Edu object, and queue it for sending

        Args:
            destination: name of server to send to
            edu_type: type of EDU to send
            content: content of EDU
            key: clobbering key for this edu
        """
        if self.is_mine_server_name(destination):
            logger.info("Not sending EDU to ourselves")
            return

        if not self._federation_shard_config.should_handle(
            self._instance_name, destination
        ):
            return

        edu = Edu(
            origin=self.server_name,
            destination=destination,
            edu_type=edu_type,
            content=content,
        )

        self.send_edu(edu, key)

    def send_edu(self, edu: Edu, key: Hashable | None) -> None:
        """Queue an EDU for sending

        Args:
            edu: edu to send
            key: clobbering key for this edu
        """
        if not self._federation_shard_config.should_handle(
            self._instance_name, edu.destination
        ):
            return

        queue = self._get_per_destination_queue(edu.destination)
        if queue is None:
            return
        if key:
            queue.send_keyed_edu(edu, key)
        else:
            queue.send_edu(edu)

    async def send_device_messages(
        self, destinations: StrCollection, immediate: bool = True
    ) -> None:
        destinations = await filter_destinations_by_retry_limiter(
            [
                destination
                for destination in destinations
                if self._federation_shard_config.should_handle(
                    self._instance_name, destination
                )
                and not self.is_mine_server_name(destination)
            ],
            clock=self.clock,
            store=self.store,
            retry_due_within_ms=CATCHUP_RETRY_INTERVAL,
        )

        for destination in destinations:
            if immediate:
                queue = self._get_per_destination_queue(destination)
                if queue is None:
                    continue
                queue.attempt_new_transaction()
            else:
                queue = self._get_per_destination_queue(destination)
                if queue is None:
                    continue
                queue.mark_new_data()
                self._destination_wakeup_queue.add_to_queue(destination)

    def wake_destination(self, destination: str) -> None:
        """Called when we want to retry sending transactions to a remote.

        This is mainly useful if the remote server has been down and we think it
        might have come back.
        """

        if self.is_mine_server_name(destination):
            logger.warning("Not waking up ourselves")
            return

        if not self._federation_shard_config.should_handle(
            self._instance_name, destination
        ):
            return

        queue = self._get_per_destination_queue(destination)
        if queue is not None:
            queue.attempt_new_transaction()

    @staticmethod
    def get_current_token() -> int:
        # Dummy implementation for case where federation sender isn't offloaded
        # to a worker.
        return 0

    def federation_ack(self, instance_name: str, token: int) -> None:
        # It is not expected that this gets called on FederationSender.
        raise NotImplementedError()

    @staticmethod
    async def get_replication_rows(
        instance_name: str, from_token: int, to_token: int, target_row_count: int
    ) -> tuple[list[tuple[int, tuple]], int, bool]:
        # Dummy implementation for case where federation sender isn't offloaded
        # to a worker.
        return [], 0, False

    async def _wake_destinations_needing_catchup(self) -> None:
        """
        Wakes up destinations that need catch-up and are not currently being
        backed off from.

        In order to reduce load spikes, adds a delay between each destination.
        """

        last_processed: str | None = None

        while not self._is_shutdown:
            destinations_to_wake = (
                await self.store.get_catch_up_outstanding_destinations(last_processed)
            )

            if not destinations_to_wake:
                # finished waking all destinations!
                break

            last_processed = destinations_to_wake[-1]

            destinations_to_wake = [
                d
                for d in destinations_to_wake
                if self._federation_shard_config.should_handle(self._instance_name, d)
                and self.hs.config.federation.is_domain_allowed_according_to_federation_whitelist(
                    d
                )
            ]

            for destination in destinations_to_wake:
                logger.info(
                    "Destination %s has outstanding catch-up, waking up.",
                    last_processed,
                )
                self.wake_destination(destination)
                await self.clock.sleep(WAKEUP_INTERVAL_BETWEEN_DESTINATIONS)



#
# This file is licensed under the Affero General Public License (AGPL) version 3.
#
# Copyright (C) 2025 New Vector, Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# See the GNU Affero General Public License for more details:
# <https://www.gnu.org/licenses/agpl-3.0.html>.
#
#

from typing import Any

from pydantic import (
    AnyHttpUrl,
    Field,
    FilePath,
    StrictBool,
    StrictStr,
    ValidationError,
    model_validator,
)
from typing_extensions import Self

from synapse.config.experimental import read_secret_from_file_once
from synapse.types import JsonDict
from synapse.util.pydantic_models import ParseModel

from ._base import Config, ConfigError, RootConfig


class MasConfigModel(ParseModel):
    enabled: StrictBool = False
    endpoint: AnyHttpUrl = AnyHttpUrl("http://localhost:8080")
    force_http2: StrictBool = False
    secret: StrictStr | None = Field(default=None)
    # We set `strict=False` to allow `str` instances.
    secret_path: FilePath | None = Field(default=None, strict=False)

    @model_validator(mode="after")
    def verify_secret(self) -> Self:
        if not self.enabled:
            return self
        if not self.secret and not self.secret_path:
            raise ValueError(
                "You must set a `secret` or `secret_path` when enabling the Matrix "
                "Authentication Service integration."
            )
        if self.secret and self.secret_path:
            raise ValueError(
                "`secret` and `secret_path` cannot be set at the same time."
            )
        return self


class MasConfig(Config):
    section = "mas"

    def read_config(
        self, config: JsonDict, allow_secrets_in_config: bool, **kwargs: Any
    ) -> None:
        mas_config = config.get("matrix_authentication_service", {})
        if mas_config is None:
            mas_config = {}

        try:
            parsed = MasConfigModel(**mas_config)
        except ValidationError as e:
            raise ConfigError(
                "Could not validate Matrix Authentication Service configuration",
                path=("matrix_authentication_service",),
            ) from e

        if parsed.secret and not allow_secrets_in_config:
            raise ConfigError(
                "Config options that expect an in-line secret as value are disabled",
                ("matrix_authentication_service", "secret"),
            )

        self.enabled = parsed.enabled
        self.endpoint = parsed.endpoint
        self.force_http2 = parsed.force_http2
        self._secret = parsed.secret
        self._secret_path = parsed.secret_path

        self.check_config_conflicts(self.root)

    def check_config_conflicts(
        self,
        root: RootConfig,
    ) -> None:
        """Checks for any configuration conflicts with other parts of Synapse.

        Raises:
            ConfigError: If there are any configuration conflicts.
        """

        if not self.enabled:
            return

        if root.experimental.msc3861.enabled:
            raise ConfigError(
                "Experimental MSC3861 was replaced by Matrix Authentication Service."
                "Please disable MSC3861 or disable Matrix Authentication Service.",
                ("experimental", "msc3861"),
            )

        if (
            root.auth.password_enabled_for_reauth
            or root.auth.password_enabled_for_login
        ):
            raise ConfigError(
                "Password auth cannot be enabled when OAuth delegation is enabled",
                ("password_config", "enabled"),
            )

        if root.registration.enable_registration:
            raise ConfigError(
                "Registration cannot be enabled when OAuth delegation is enabled",
                ("enable_registration",),
            )

        # We only need to test the user consent version, as if it must be set if the user_consent section was present in the config
        if root.consent.user_consent_version is not None:
            raise ConfigError(
                "User consent cannot be enabled when OAuth delegation is enabled",
                ("user_consent",),
            )

        if (
            root.oidc.oidc_enabled
            or root.saml2.saml2_enabled
            or root.cas.cas_enabled
            or root.jwt.jwt_enabled
        ):
            raise ConfigError("SSO cannot be enabled when OAuth delegation is enabled")

        if bool(root.authproviders.password_providers):
            raise ConfigError(
                "Password auth providers cannot be enabled when OAuth delegation is enabled"
            )

        if root.captcha.enable_registration_captcha:
            raise ConfigError(
                "CAPTCHA cannot be enabled when OAuth delegation is enabled",
                ("captcha", "enable_registration_captcha"),
            )

        if root.auth.login_via_existing_enabled:
            raise ConfigError(
                "Login via existing session cannot be enabled when OAuth delegation is enabled",
                ("login_via_existing_session", "enabled"),
            )

        if root.registration.refresh_token_lifetime:
            raise ConfigError(
                "refresh_token_lifetime cannot be set when OAuth delegation is enabled",
                ("refresh_token_lifetime",),
            )

        if root.registration.nonrefreshable_access_token_lifetime:
            raise ConfigError(
                "nonrefreshable_access_token_lifetime cannot be set when OAuth delegation is enabled",
                ("nonrefreshable_access_token_lifetime",),
            )

        if root.registration.session_lifetime:
            raise ConfigError(
                "session_lifetime cannot be set when OAuth delegation is enabled",
                ("session_lifetime",),
            )

        if root.registration.enable_3pid_changes:
            raise ConfigError(
                "enable_3pid_changes cannot be enabled when OAuth delegation is enabled",
                ("enable_3pid_changes",),
            )

    def secret(self) -> str:
        if self._secret is not None:
            return self._secret
        elif self._secret_path is not None:
            return read_secret_from_file_once(
                str(self._secret_path),
                ("matrix_authentication_service", "secret_path"),
            )
        else:
            raise RuntimeError(
                "Neither `secret` nor `secret_path` are set, this is a bug.",
            )
